import unittest
from typing import List
from unittest.mock import patch
import datetime as dt
import polars as pl

from xml_common.utils import type_to_pl

orig_import = __import__

def import_mock(name, globals=None, locals=None, fromlist=(), level=0):
    if name == 'polars':
        raise ImportError(f"No module named '{name}'")
    return orig_import(name, globals, locals, fromlist, level)

class TestType:
    pass


class PlFormatTest(unittest.TestCase):
    @patch('builtins.__import__', side_effect=import_mock)
    def test_import_error(self, mock_import):
        self.assertRaisesRegex(ImportError, "polars not installed", type_to_pl, int)

    def test_int(self):
        self.assertEqual(pl.Int64, type_to_pl(int), "Incorrect type (int) to polars type.")

    def test_float(self):
        self.assertEqual(pl.Float64, type_to_pl(float), "Incorrect type (float) to polars type.")

    def test_bool(self):
        self.assertEqual(pl.Boolean, type_to_pl(bool), "Incorrect type (bool) to polars type.")

    def test_datetime(self):
        self.assertEqual(pl.Datetime, type_to_pl(dt.datetime), "Incorrect type (datatime) to polars type.")

    def test_str(self):
        self.assertEqual(pl.String, type_to_pl(str), "Incorrect type (str) to polars type.")

    def test_list(self):
        self.assertEqual(pl.List(pl.String), type_to_pl(List[str]), "Incorrect type (List[str]) to polars type.")
        self.assertEqual(pl.List(pl.Object), type_to_pl(List[TestType]), "Incorrect type (List[obj]) to polars type.")
        self.assertEqual(pl.List(pl.Object), type_to_pl(List), "Incorrect type (List[obj]) to polars type.")
        self.assertEqual(pl.String, type_to_pl(List[str], lazy=True), "Incorrect type [lazy] (List[str]) to polars type.")

    def test_nested_list(self):
        self.assertEqual(pl.List(pl.List(pl.String)), type_to_pl(List[List[str]]), "Incorrect type (List[List[str]]) to polars type.")
        self.assertEqual(pl.List(pl.List(pl.Object)), type_to_pl(List[List[TestType]]), "Incorrect type (List[List[obj]]) to polars type.")
        self.assertEqual(pl.List(pl.List(pl.List(pl.Int64))), type_to_pl(List[List[List[int]]]),
                         "Incorrect type (List[List[List[int]]]) to polars type.")
        self.assertEqual(pl.String, type_to_pl(List[List[List[int]]], lazy=True),
                         "Incorrect type [lazy] (List[List[List[int]]]) to polars type.")

    def test_obj(self):
        self.assertEqual(pl.Object, type_to_pl(TestType), "Incorrect type (obj) to polars type.")
        self.assertEqual(pl.String, type_to_pl(TestType, lazy=True), "Incorrect type [lazy] (obj) to polars type.")



if __name__ == '__main__':
    unittest.main()



import unittest

from lxml import etree as et

from eml.resources import EMLCoverage
from test_xml.test_xml import TestXML


class TestCoverage(TestXML):
    DEFAULT_TAGS = {
        "scope": "document",
    }

    def test_parse_none(self):
        self.assertIsNone(EMLCoverage.parse(None, {}), "Distribution from nowhere")

    def test_parse(self):
        xml_text = """
<coverage id="1" scope="document">
    <geographicCoverage>
        <references>2</references>
    </geographicCoverage>
    <temporalCoverage>
        <references>3</references>
    </temporalCoverage>
    <taxonomicCoverage>
        <references>4</references>
    </taxonomicCoverage>
</coverage>
        """
        coverage = EMLCoverage.from_string(xml_text)
        self.assertEqual("1", coverage.id, "Error on parsing id")
        self.assertEqual(
            "2",
            coverage.geographic.id,
            "Error on parsing geographic Coverage"
        )
        self.assertEqual(
            "3",
            coverage.temporal.id,
            "Error on parsing geographic Coverage"
        )
        self.assertEqual(
            "4",
            coverage.taxonomic.id,
            "Error on parsing geographic Coverage"
        )
        self.assertFalse(
            coverage.referencing,
            "Error on parsing referencing"
        )
        self.assertEqualTree(et.fromstring(xml_text), coverage.to_element(), "Creator error on to element")

    def test_parse_referencing(self):
        xml_text = """
<coverage>
    <references>1</references>
</coverage>
        """
        coverage = EMLCoverage.from_string(xml_text)
        self.assertEqual("1", coverage.id, "Error on parsing id")
        self.assertTrue(
            coverage.referencing,
            "Error on parsing referencing"
        )
        self.assertEqualTree(et.fromstring(xml_text), coverage.to_element(), "Creator error on to element")

    def test_parse_invalid(self):
        self.assertRaises(
            TypeError,
            EMLCoverage.from_string,
            "<coverage></coverage>"
        )


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
import datetime as dt
from typing import Dict
from lxml import etree as et
from xml_common import XMLObject
from xml_common.utils import format_datetime
from xml_common.utils.type_functions import unformat_datetime


class ChangeHistory(XMLObject):
    """
    Attributes:
    -----------
    change_scope : str
        An expression describing the scope to which the documented change was applied.
    old_value : str
        The previous value of the data prior to the change.
    change_date : datetime.datetime
        The date the changes were applied.
    comment : str, optional
        Explanation or justification for the change made to the data.
    """
    PRINCIPAL_TAG = "changeHistory"

    def __init__(
            self, change_scope: str, old_value: str,
            change_date: dt.date, comment: str = None
    ) -> None:
        super().__init__()
        self.__change_scope__ = change_scope
        self.__old_value__ = old_value
        self.__change_dt__ = change_date
        self.__comment__ = comment
        return

    @property
    def change_scope(self) -> str:
        """str: An expression describing the scope to which the documented change was applied."""
        return self.__change_scope__

    @property
    def old_value(self) -> str:
        """str: The previous value of the data prior to the change."""
        return self.__old_value__

    @property
    def change_date(self) -> dt.date:
        """datetime.date: The date the changes were applied."""
        return self.__change_dt__

    @property
    def comment(self) -> str:
        """str: Explanation or justification for the change made to the data."""
        return self.__comment__

    @classmethod
    def parse(cls, element: et.Element, nmap: Dict) -> ChangeHistory | None:
        """
        Parses a lxml element into a ChangeHistory object.

        Parameters
        ----------
        element : lxml.tree.Element
            Object in the Element format to be parsed into a Python object.
        nmap : Dict
            Dictionary of namespace.

        Returns
        -------
        ChangeHistory
            Instance from parsed xml element.
        """
        if element is None:
            return None
        comment_elem = element.find("comment", nmap)
        comment = comment_elem.text if comment_elem is not None else None
        return ChangeHistory(
            change_scope=element.find("changeScope", nmap).text,
            old_value=element.find("oldValue", nmap).text,
            change_date=format_datetime(element.find("changeDate", nmap).text).date(),
            comment=comment
        )

    def to_element(self) -> et.Element:
        """
        Generate an XML element with ChangeHistory instance information.

        Returns
        -------
        lxml.etree.Element
            XML element object
        """
        change_hist_elem = super().to_element()
        change_scope_elem = self.object_to_element("changeScope")
        change_scope_elem.text = self.change_scope
        change_hist_elem.append(change_scope_elem)
        old_value_elem = self.object_to_element("oldValue")
        old_value_elem.text = self.old_value
        change_hist_elem.append(old_value_elem)
        change_date_elem = self.object_to_element("changeDate")
        change_date_elem.text = unformat_datetime(dt.datetime.combine(self.change_date, dt.time()))
        change_hist_elem.append(change_date_elem)
        if self.comment is not None:
            comment_elem = self.object_to_element("comment")
            comment_elem.text = self.comment
            change_hist_elem.append(comment_elem)
        return change_hist_elem



from __future__ import annotations

from abc import ABC, abstractmethod


class DarwinCore(ABC):
    """
    Base class of this package.
    """
    def __init__(self) -> None:
        return

    @classmethod
    @abstractmethod
    def from_file(cls, path_to_archive: str) -> DarwinCore:
        """
        Generate a Darwin Core Standard from a file.

        Parameters
        ----------
        path_to_archive : str
            Path of the archive file.

        Returns
        -------
        DarwinCore
            Instance of the Darwin Core Standard.
        """
        pass

    @abstractmethod
    def to_file(self, path_to_archive: str, encoding: str = "utf-8") -> None:
        """
        Generate a Darwin Core file using the information of this instance.

        Parameters
        ----------
        path_to_archive : str
            Path of the archive to generate.
        encoding : str, optional
            Encoding of the corresponding files. Default `"utf-8"`.
        """
        pass



import unittest
from lxml import etree as et

from xml_common.utils import Language
from eml.types import I18nString, ExtensionString
from test_xml.test_xml import TestXML


class TestI18nString(TestXML):
    def test_i18n_string(self):
        variable = I18nString("My variable")
        self.assertEqual("My variable", variable, "Not equal to string")
        self.assertEqual("My variable", str(variable), "Not equal to string")
        self.assertEqual("<i18n String (My variable) [eng]>", repr(variable), "Not equal to string")
        self.assertEqual(Language.ENG, variable.language, "Language set incorrectly")

    def test_i18n_spanish(self):
        variable = I18nString("Mi variable", lang=Language.ESP)
        self.assertEqual("Mi variable", variable, "Not equal to string")
        self.assertEqual("Mi variable", str(variable), "Not equal to string")
        self.assertEqual("<i18n String (Mi variable) [esp]>", repr(variable), "Not equal to string")
        self.assertEqual(Language.ESP, variable.language, "Language set incorrectly")

    def test_parse(self):
        self.assertIsNone(I18nString.parse(None, {}), "Parsing something")
        example_xml = """
<title
    xml:lang="eng"
>My variable</title>
        """
        variable = I18nString.from_string(example_xml)
        self.assertEqual("My variable", variable, "Not equal to string")
        self.assertEqual(Language.ENG, variable.language, "Language set incorrectly")

    def test_wrong_language_type(self):
        self.assertRaises(
            TypeError, I18nString, "My variable", 20
        )

    def test_to_element(self):
        variable = I18nString("Mi variable", lang=Language.ESP)
        self.assertRaises(RuntimeError, variable.to_element)
        variable.set_tag("title")
        expected = et.Element("title")
        expected.text = "Mi variable"
        expected.set("{http://www.w3.org/XML/1998/namespace}lang", "esp")
        self.assertEqualTree(expected, variable.to_element(), "Error on element")

    def test_equal(self):
        variable = I18nString("My variable")
        self.assertEqual(I18nString("My variable"), variable, "Error on equal implementation")
        self.assertNotEqual(
            I18nString("My variable", lang=Language.ESP),
            variable,
            "Error on equal implementation"
        )

    def test_not_equal(self):
        variable = I18nString("My variable")
        self.assertNotEqual(ExtensionString("My variable"), variable, "Error on equal implementation")
        self.assertNotEqual(I18nString("Another variable"), variable, "Error on equal implementation")
        self.assertNotEqual(30, variable, "Error on equal implementation")

    def test_wrong_language(self):
        self.assertRaises(
            NotImplementedError, I18nString, "My variable", "6f5"
        )
        self.assertRaises(
            NotImplementedError, I18nString, "My variable", "de"
        )
        self.assertRaises(
            NotImplementedError, I18nString, "My variable", "fr"
        )

    def test_sort(self):
        first_variable = I18nString("A variable", Language.ENG)
        second_variable = I18nString("Mi variable", Language.ESP)
        self.assertGreater(second_variable, first_variable, "'A' greater than 'M'")
        self.assertGreaterEqual(second_variable, first_variable, "'A' greater equal than 'M'")
        self.assertGreater(second_variable, "A variable", "'A' greater than 'M' (on I18nString and str)")
        self.assertGreaterEqual(second_variable, "A variable", "'A' greater equal than 'M' (on I18nString and str)")
        self.assertRaises(TypeError, first_variable.__lt__, 45)
        self.assertRaises(TypeError, first_variable.__le__, 45)
        self.assertLess(first_variable, second_variable, "'M' less than 'A'")
        self.assertLessEqual(first_variable, second_variable, "'M' less equal than 'A'")
        self.assertLess(first_variable, "Mi variable", "'M' less than 'A' (on I18nString and str)")
        self.assertLessEqual(first_variable, "Mi variable", "'M' less equal than 'A' (on I18nString and str)")
        self.assertRaises(TypeError, first_variable.__gt__, 45)
        self.assertRaises(TypeError, first_variable.__ge__, 45)


if __name__ == '__main__':
    unittest.main()



import os
import unittest
from lxml import etree as et

from xml_common.utils import Language
from eml.resources import EMLDataset
from eml.types import Scope, SemanticAnnotation
from test_xml.test_xml import TestXML

PATH = os.path.abspath(os.path.dirname(__file__))


class TestEMLDataset(TestXML):
    DEFAULT_TAGS = {
        "scope": "document",
        "{http://www.w3.org/XML/1998/namespace}lang": "eng",
        "phonetype": "voice",
    }

    def setUp(self) -> None:
        with open(os.path.join(PATH, os.pardir, "example_data", "eml.xml"), "r", encoding="utf-8") as file:
            content = file.read()
        base_file = et.fromstring(content)
        self.dataset_xml = base_file.find("dataset", namespaces=base_file.nsmap)
        self.text = et.tostring(self.dataset_xml, pretty_print=True).decode("utf-8")
        return

    def test_parse(self):
        dataset = EMLDataset.from_string(self.text)
        self.assertEqual(
            """\tTitle: Example for Darwin Core Archive
\tCreator: Creator Organization
\tMetadataProvider: Metadata Manager at Metadata Provider Organization
\tCustodian Steward: Doe, J. (Custodian Steward)
\tOriginator: Doe, J.
\tAuthor: Example, J.""",
            str(dataset),
            "Error on parse to string"
        )
        self.assertIsNone(dataset.id, "Id from nothing")
        self.assertIsNone(dataset.system, "System from nothing")
        self.assertEqual(Scope.DOCUMENT, dataset.scope, "Not given different than default")
        self.assertFalse(dataset.referencing, "Reference from nothing")
        self.assertEqual(1, len(dataset.titles), "Extra titles from nowhere")
        self.assertEqualTree(self.dataset_xml, dataset.to_element(), "Wrong element generated")
        return

    def test_missing_title(self):
        dataset_xml = """
<dataset>
    <alternateIdentifier system="http://gbif.org">VCR3465</alternateIdentifier>
    <shortName>A short description</shortName>
</dataset>
        """
        self.assertRaises(
            ValueError,
            EMLDataset.from_string,
            dataset_xml
        )

    def test_more_titles(self):
        dataset_xml = """
<dataset>
    <title>A Title</title>
    <title xml:lang="esp">Un Título</title>
    <title xml:lang="esp">Un Título Diferente</title>
    <creator>
        <organizationName>Organization Creator</organizationName>
    </creator>
    <alternateIdentifier system="http://gbif.org">VCR3465</alternateIdentifier>
    <shortName>A short description</shortName>
    <contact>
        <organizationName>Organization Contact</organizationName>
    </contact>
</dataset>
        """
        eml_dataset = EMLDataset.from_string(dataset_xml)
        self.assertNotEqual("Un Título", eml_dataset.title, "Title set in wrong order.")
        self.assertEqual("A Title", eml_dataset.title, "Title not set.")
        self.assertEqual("Un Título Diferente", eml_dataset.titles[2], "Extra titles set in wrong order.")
        self.assertEqual(Language.ENG, eml_dataset.title.language, "Title set in wrong language.")
        self.assertEqualTree(et.fromstring(dataset_xml), eml_dataset.to_element(), "Wrong to element.")

    def test_alternate_identifier(self):
        dataset_xml = """
<dataset>
    <title xml:lang="eng">Test Title</title>
    <creator>
        <organizationName>Organization Creator</organizationName>
    </creator>
    <alternateIdentifier system="http://gbif.org">VCR3465</alternateIdentifier>
    <shortName>A short description</shortName>
    <contact>
        <organizationName>Organization Contact</organizationName>
    </contact>
</dataset>
        """
        eml_dataset = EMLDataset.from_string(dataset_xml)
        self.assertEqual("VCR3465", eml_dataset.alternative_identifiers[0], "Alternative id set incorrectly")
        self.assertEqual("http://gbif.org", eml_dataset.alternative_identifiers[0].system,
                         "Alternative id set incorrectly")
        self.assertEqual("A short description", eml_dataset.short_name,
                         "Short name not set correctly")
        self.assertEqualTree(
            et.fromstring(dataset_xml),
            eml_dataset.to_element(),
            "Wrong alternative identifier to element"
        )

    def test_alternate_identifiers(self):
        dataset_xml = """
<dataset>
    <title>Title</title>
    <creator>
        <organizationName>Organization Creator</organizationName>
    </creator>
    <alternateIdentifier system="http://gbif.org">VCR3465</alternateIdentifier>
    <alternateIdentifier>VCR3465</alternateIdentifier>
    <contact>
        <organizationName>Organization Contact</organizationName>
    </contact>
</dataset>
        """
        eml_dataset = EMLDataset.from_string(dataset_xml)
        self.assertEqual("VCR3465", eml_dataset.alternative_identifiers[0], "Alternative id set incorrectly")
        self.assertEqual("http://gbif.org", eml_dataset.alternative_identifiers[0].system,
                         "Alternative id set incorrectly")
        self.assertEqual("VCR3465", eml_dataset.alternative_identifiers[1], "Alternative system set incorrectly")
        self.assertIsNone(eml_dataset.alternative_identifiers[1].system, "Alternative system set incorrectly")
        self.assertEqualTree(
            et.fromstring(dataset_xml),
            eml_dataset.to_element(),
            "Wrong alternative identifiers to element"
        )

    def test_reference_dataset(self):
        reference_dataset_xml = """
<dataset>
    <references>1</references>
</dataset>
        """
        eml_dataset = EMLDataset.from_string(reference_dataset_xml)
        self.assertEqual("1", eml_dataset.id, "Id set incorrectly")
        self.assertTrue(eml_dataset.referencing, "Referencing wrong initialized")
        self.assertIsNone(
            eml_dataset.references.system,
            "References system wrong initialized"
        )
        self.assertEqualTree(
            et.fromstring(reference_dataset_xml),
            eml_dataset.to_element(),
            "Wrong referrer to element"
        )

    def test_annotate_invalid(self):
        dataset = EMLDataset.from_string(self.text)
        self.assertEqual(0, len(dataset.annotation), "Annotation from nowhere")
        self.assertRaises(
            ValueError,
            dataset.annotate,
            SemanticAnnotation(
                (
                    "http://ecoinformatics.org/oboe/oboe.1.2/oboe-characteristics.owl#ofCharacteristic",
                    "has characteristic"
                ),(
                    "http://ecoinformatics.org/oboe/oboe.1.2/oboe-characteristics.owl#Mass",
                    "Mass"
                )
            )
        )

    def test_annotate(self):
        dataset = EMLDataset.from_string(self.text)
        dataset.__id__ = "1"
        self.assertEqual(0, len(dataset.annotation), "Annotation from nowhere")
        annotation = SemanticAnnotation(
            ("http://ecoinformatics.org/oboe/oboe.1.2/oboe-characteristics.owl#ofCharacteristic", "has characteristic"),
            ("http://ecoinformatics.org/oboe/oboe.1.2/oboe-characteristics.owl#Mass", "Mass")
        )
        dataset.annotate(annotation)
        self.assertEqual(1, len(dataset.annotation), "Annotation not set")
        text_as_element = et.fromstring(self.text)
        annotation.set_tag("annotation")
        text_as_element.set("id", "1")
        text_as_element.append(annotation.to_element())
        self.assertEqualTree(text_as_element, dataset.to_element(), "Annotation not set in element tree")

    def test_reference_system(self):
        reference_dataset_xml = """
<dataset>
    <references system="http://gbif.org">1</references>
</dataset>
        """
        eml_dataset = EMLDataset.from_string(reference_dataset_xml)
        self.assertEqual("1", eml_dataset.id, "Id set incorrectly")
        self.assertTrue(eml_dataset.referencing, "Referencing wrong initialized")
        self.assertEqual(
            "http://gbif.org",
            eml_dataset.references.system,
            "References system wrong initialized"
        )
        self.assertEqualTree(
            et.fromstring(reference_dataset_xml),
            eml_dataset.to_element(),
            "Wrong referrer to element"
        )

    def test_invalid_reference(self):
        reference_dataset_xml = """
<dataset id="10">
    <references>1</references>
</dataset>
        """
        self.assertRaises(
            AssertionError,
            EMLDataset.from_string,
            reference_dataset_xml
        )

    def test_no_contact(self):
        dataset_xml = """
<dataset>
    <title>A Title</title>
    <creator>
        <organizationName>Organization Creator</organizationName>
    </creator>
    <alternateIdentifier system="http://gbif.org">VCR3465</alternateIdentifier>
    <shortName>A short description</shortName>
</dataset>
        """
        with self.assertRaisesRegex(ValueError, "No contact provided."):
            EMLDataset.from_string(dataset_xml)
        return

    def test_purpose(self):
        dataset_xml = """
<dataset>
    <title>A Title</title>
    <creator>
        <organizationName>Organization Creator</organizationName>
    </creator>
    <alternateIdentifier system="http://gbif.org">VCR3465</alternateIdentifier>
    <shortName>A short description</shortName>
    <purpose>
        <para>Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nam nec.</para>
    </purpose>
    <contact>
        <organizationName>Organization Contact</organizationName>
    </contact>
</dataset>
        """
        eml_dataset = EMLDataset.from_string(dataset_xml)
        self.assertNotEqual("Un Título", eml_dataset.title, "Title set in wrong order.")
        self.assertEqual(Language.ENG, eml_dataset.title.language, "Title set in wrong language,")
        self.assertEqual("Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nam nec.", eml_dataset.purpose.paragraphs[0], "Purpose not parsed.")
        self.assertEqualTree(et.fromstring(dataset_xml), eml_dataset.to_element(), "Wrong to element.")

    def test_intro(self):
        dataset_xml = """
<dataset>
    <title>A Title</title>
    <creator>
        <organizationName>Organization Creator</organizationName>
    </creator>
    <alternateIdentifier system="http://gbif.org">VCR3465</alternateIdentifier>
    <shortName>A short description</shortName>
    <introduction>
        <para>Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nam nec.</para>
    </introduction>
    <contact>
        <organizationName>Organization Contact</organizationName>
    </contact>
</dataset>
        """
        eml_dataset = EMLDataset.from_string(dataset_xml)
        self.assertNotEqual("Un Título", eml_dataset.title, "Title set in wrong order.")
        self.assertEqual(Language.ENG, eml_dataset.title.language, "Title set in wrong language,")
        self.assertEqual("Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nam nec.",
                         eml_dataset.introduction.paragraphs[0], "Introduction not parsed.")
        self.assertEqualTree(et.fromstring(dataset_xml), eml_dataset.to_element(), "Wrong to element.")

    def test_getting_started(self):
        dataset_xml = """
<dataset>
    <title>A Title</title>
    <creator>
        <organizationName>Organization Creator</organizationName>
    </creator>
    <alternateIdentifier system="http://gbif.org">VCR3465</alternateIdentifier>
    <shortName>A short description</shortName>
    <gettingStarted>
        <para>Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nam nec.</para>
    </gettingStarted>
    <contact>
        <organizationName>Organization Contact</organizationName>
    </contact>
</dataset>
        """
        eml_dataset = EMLDataset.from_string(dataset_xml)
        self.assertNotEqual("Un Título", eml_dataset.title, "Title set in wrong order.")
        self.assertEqual(Language.ENG, eml_dataset.title.language, "Title set in wrong language,")
        self.assertEqual("Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nam nec.",
                         eml_dataset.getting_started.paragraphs[0], "Getting Started not parsed.")
        self.assertEqualTree(et.fromstring(dataset_xml), eml_dataset.to_element(), "Wrong to element.")

    def test_acknowledgements(self):
        dataset_xml = """
<dataset>
    <title>A Title</title>
    <creator>
        <organizationName>Organization Creator</organizationName>
    </creator>
    <alternateIdentifier system="http://gbif.org">VCR3465</alternateIdentifier>
    <shortName>A short description</shortName>
    <acknowledgements>
        <para>Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nam nec.</para>
    </acknowledgements>
    <contact>
        <organizationName>Organization Contact</organizationName>
    </contact>
</dataset>
        """
        eml_dataset = EMLDataset.from_string(dataset_xml)
        self.assertNotEqual("Un Título", eml_dataset.title, "Title set in wrong order.")
        self.assertEqual(Language.ENG, eml_dataset.title.language, "Title set in wrong language,")
        self.assertEqual("Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nam nec.",
                         eml_dataset.acknowledgements.paragraphs[0], "Acknowledgements not parsed.")
        self.assertEqualTree(et.fromstring(dataset_xml), eml_dataset.to_element(), "Wrong to element.")

    def test_maintenance(self):
        dataset_xml = """
<dataset>
    <title>A Title</title>
    <creator>
        <organizationName>Organization Creator</organizationName>
    </creator>
    <alternateIdentifier system="http://gbif.org">VCR3465</alternateIdentifier>
    <shortName>A short description</shortName>
    <maintenance>
        <description>
            <para>Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nam nec.</para>
        </description>
    </maintenance>
    <contact>
        <organizationName>Organization Contact</organizationName>
    </contact>
</dataset>
        """
        eml_dataset = EMLDataset.from_string(dataset_xml)
        self.assertNotEqual("Un Título", eml_dataset.title, "Title set in wrong order.")
        self.assertEqual(Language.ENG, eml_dataset.title.language, "Title set in wrong language,")
        self.assertEqual("Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nam nec.",
                         eml_dataset.maintenance.description.paragraphs[0], "Maintenance not parsed.")
        self.assertEqualTree(et.fromstring(dataset_xml), eml_dataset.to_element(), "Wrong to element.")

    def test_publisher(self):
        dataset_xml = """
<dataset>
    <title>A Title</title>
    <creator>
        <organizationName>Organization Creator</organizationName>
    </creator>
    <alternateIdentifier system="http://gbif.org">VCR3465</alternateIdentifier>
    <shortName>A short description</shortName>
    <contact>
        <organizationName>Organization Contact</organizationName>
    </contact>
    <publisher>
        <organizationName>Organization Publisher</organizationName>
    </publisher>
</dataset>
        """
        eml_dataset = EMLDataset.from_string(dataset_xml)
        self.assertNotEqual("Un Título", eml_dataset.title, "Title set in wrong order.")
        self.assertEqual(Language.ENG, eml_dataset.title.language, "Title set in wrong language,")
        self.assertEqual("Organization Publisher",
                         str(eml_dataset.publisher.organization_name),
                         "Publisher not parsed.")
        self.assertEqualTree(et.fromstring(dataset_xml), eml_dataset.to_element(), "Wrong to element.")

    def test_pub_place(self):
        dataset_xml = """
<dataset>
    <title>A Title</title>
    <creator>
        <organizationName>Organization Creator</organizationName>
    </creator>
    <alternateIdentifier system="http://gbif.org">VCR3465</alternateIdentifier>
    <shortName>A short description</shortName>
    <contact>
        <organizationName>Organization Contact</organizationName>
    </contact>
    <pubPlace>San Francisco, CA, USA</pubPlace>
</dataset>
        """
        eml_dataset = EMLDataset.from_string(dataset_xml)
        self.assertNotEqual("Un Título", eml_dataset.title, "Title set in wrong order.")
        self.assertEqual(Language.ENG, eml_dataset.title.language, "Title set in wrong language,")
        self.assertEqual("San Francisco, CA, USA",
                         str(eml_dataset.pub_place),
                         "Publication place not parsed.")
        self.assertEqualTree(et.fromstring(dataset_xml), eml_dataset.to_element(), "Wrong to element.")

    def test_methods(self):
        dataset_xml = """
<dataset>
    <title>A Title</title>
    <creator>
        <organizationName>Organization Creator</organizationName>
    </creator>
    <alternateIdentifier system="http://gbif.org">VCR3465</alternateIdentifier>
    <shortName>A short description</shortName>
    <contact>
        <organizationName>Organization Contact</organizationName>
    </contact>
    <methods>
        <methodStep>
            <description>
                <para>Lorem ipsum dolor sit amet, consectetur adipiscing elit. Etiam vitae diam nec eros interdum posuere.</para>
            </description>
        </methodStep>
    </methods>
</dataset>
        """
        eml_dataset = EMLDataset.from_string(dataset_xml)
        self.assertNotEqual("Un Título", eml_dataset.title, "Title set in wrong order.")
        self.assertEqual(Language.ENG, eml_dataset.title.language, "Title set in wrong language,")
        self.assertEqual(
            "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Etiam vitae diam nec eros interdum posuere.",
            eml_dataset.methods.method_steps[0].description.paragraphs[0],
            "Methods not parsed."
        )
        self.assertEqualTree(et.fromstring(dataset_xml), eml_dataset.to_element(), "Wrong to element.")

    def test_project(self):
        dataset_xml = """
<dataset>
    <title>A Title</title>
    <creator>
        <organizationName>Organization Creator</organizationName>
    </creator>
    <alternateIdentifier system="http://gbif.org">VCR3465</alternateIdentifier>
    <shortName>A short description</shortName>
    <contact>
        <organizationName>Organization Contact</organizationName>
    </contact>
    <project>
        <title>Example Title</title>
        <personnel>
            <organizationName>Example Organization</organizationName>
            <role>contentProvider</role>
        </personnel>
    </project>
</dataset>
        """
        eml_dataset = EMLDataset.from_string(dataset_xml)
        self.assertNotEqual("Un Título", eml_dataset.title, "Title set in wrong order.")
        self.assertEqual(Language.ENG, eml_dataset.title.language, "Title set in wrong language,")
        self.assertEqual(
            "Example Title",
            eml_dataset.project.title,
            "Project not parsed."
        )
        self.assertEqualTree(et.fromstring(dataset_xml), eml_dataset.to_element(), "Wrong to element.")


if __name__ == '__main__':
    unittest.main()



from typing import List

from dwca.terms import Field


class TaxonID(Field):
    """
    An identifier for the set of Taxon information.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/taxonID"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class ScientificNameID(Field):
    """
    An identifier for the nomenclatural (not taxonomic) details of a scientific name.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/scientificNameID"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class AcceptedNameUsageID(Field):
    """
    An identifier for the name usage of the taxon.

    This term should be used for synonyms or misapplied names
    to refer to the taxonID of a Taxon record that represents
    the accepted (botanical) or valid (zoological) name. For
    Darwin Core Archives the related record should be present
    locally in the same archive.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/acceptedNameUsageID"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class ParentNameUsageID(Field):
    """
    An identifier for the name usage f the direct, most proximate higher-rank parent taxon.

    The name usage, documented meaning of the name according
    to a source, of the direct, most proximate higher-rank
    parent taxon (in a classification) of the most specific
    element of the Scientific Name.
    This term should be used for accepted names to refer to
    the taxonID of a Taxon record that represents the next
    higher taxon rank in the same taxonomic classification.
    For Darwin Core Archives the related record should be
    present locally in the same archive.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/parentNameUsageID"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class OriginalNameUsageID(Field):
    """
    An identifier for the name usage in which the terminal term was originally established.

    This term should be used to refer to the taxonID of a
    Taxon record that represents the usage of the terminal
    element of the scientificName as originally established
    under the rules of the associated nomenclaturalCode.
    For example, for names governed by the ICNafp, this
    term would establish the relationship between a record
    representing a subsequent combination and the record
    for its corresponding basionym. Unlike basionyms,
    however, this term can apply to scientific names
    at all ranks. For Darwin Core Archives the related
    record should be present locally in the same archive.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/originalNameUsageID"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class NameAccordingToID(Field):
    """
    An identifier for the source in which the specific taxon concept circumscription is defined or implied.

    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/nameAccordingToID"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class NamePublishedInID(Field):
    """
    An identifier for the publication in which the scientificName was originally established.

    Under the rules of the associated Nomenclatural Code.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/namePublishedInID"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class TaxonConceptID(Field):
    """
    An identifier for the taxonomic concept to which the record refers.

    Not for the nomenclatural details of a Taxon.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/taxonConceptID"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class ScientificName(Field):
    """
    The full scientific name, with authorship and date information if known.

    When forming part of an Identification, this should be the name
    in the lowest level taxonomic rank that can be determined.
    This term should not contain identification qualifications,
    which should instead be supplied in the identificationQualifier
    term.
    When applied to an Organism or Occurrence, this term should be
    used to represent the scientific name that was applied to the
    associated Organism in accordance with the Taxon to which it
    was or is currently identified. Names should be compliant to
    the most recent nomenclatural code. For example, names of
    hybrids for algae, fungi and plants should follow the rules
    of the International Code of Nomenclature for algae, fungi,
    and plants (Schenzhen Code Articles H.1, H.2 and H.3). Thus,
    use the multiplication sign × (Unicode U+00D7, HTML ×) to
    identify a hybrid, not x or X, if possible.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/scientificName"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class AcceptedNameUsage(Field):
    """
    The full name, with authorship and date information if known, of the Taxon.

    The full scientific name, with authorship and date
    information if known, of the accepted (botanical)
    or valid (zoological) name in cases where the
    provided Scientific Name is considered by the
    reference indicated in the ``According To`` property,
    or of the content provider, to be a synonym or misapplied
    name. When applied to an Organism or an Occurrence, this
    term should be used in cases where a content provider
    regards the provided Scientific Name to be inconsistent
    with the taxonomic perspective of the content provider.
    For example, there are many discrepancies within specimen
    collections and observation datasets between the recorded
    name (e.g., the most recent identification from an expert
    who examined a specimen, or a field identification for an
    observed Organism), and the name asserted by the content
    provider to be taxonomically accepted.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/acceptedNameUsage"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class ParentNameUsage(Field):
    """
    The direct, most proximate higher-rank parent Taxon of the most specific element of the Scientific Name.

    The full name, with authorship and date information
    if known, of the direct, most proximate higher-rank
    parent Taxon (in a classification) of the most specific
    element of the Scientific Name.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/parentNameUsage"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class OriginalNameUsage(Field):
    """
    The taxon name as it originally appeared when first established.

    The taxon name, with authorship and date information if
    known, as it originally appeared when first established
    under the rules of the associated Nomenclatural Code. The
    basionym (botany) or basonym (bacteriology) of the
    Scientific Name or the senior/earlier homonym for
    replaced names. For example, for names governed by the
    ICNafp, this term would indicate the basionym of a record
    representing a subsequent combination. Unlike basionyms,
    however, this term can apply to scientific names at all ranks.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/originalNameUsage"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class NameAccordingTo(Field):
    """
    The reference to the source in which the specific taxon concept circumscription is defined or implied.

    Traditionally signified by the Latin "sensu" or "sec."
    (from secundum, meaning "according to"). For taxa that
    result from identifications, a reference to the keys,
    monographs, experts and other sources should be given.
    This term provides context to the Scientific Name. Together
    with the Scientific Name, separated by sensu or sec., it
    forms the taxon concept label, which may be seen as having
    the same relationship to ``Taxon Concept ID`` as, for example,
    Accepted Name Usage has to ``Accepted Name Usage ID``. When
    not provided, in Taxon Core data sets the Name According To
    can be taken to be the data set. In this case the data set
    mostly provides sufficient context to infer the delimitation
    of the taxon and its relationship with other taxa. In Occurrence
    Core data sets, when not provided, Name According To can be
    an underlying taxonomy of the data set, e.g. `Plants of the
    World Online <http://powo.science.kew.org/>`_ for vascular
    plant records in iNaturalist (in which case it should be
    provided), or, which is the case for most Preserved Specimen
    data sets, the Identification, in which case there is no
    further context.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/nameAccordingTo"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class NamePublishedIn(Field):
    """
    A reference for the publication in which the Scientific Name was originally established.

    Established under the rules of the associated Nomenclatural Code.
    A citation of the first publication of the name in its given
    combination, not the basionym / original name. Recombinations
    are often not published in zoology, in which case Name Published In
    should be empty.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/namePublishedIn"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class NamePublishedInYear(Field):
    """
    The four-digit year in which the Scientific Name was published.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/namePublishedInYear"
    TYPE = int

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class HigherClassification(Field):
    """
    A list of taxa names terminating at the rank immediately superior to the referenced Taxon.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/higherClassification"
    TYPE = List[str]

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class Kingdom(Field):
    """
    The full scientific name of the kingdom in which the Taxon is classified.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/kingdom"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class Phylum(Field):
    """
    The full scientific name of the phylum or division in which the Taxon is classified.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/phylum"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class DWCClass(Field):
    """
    The full scientific name of the class in which the Taxon is classified.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/class"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class Order(Field):
    """
    The full scientific name of the order in which the Taxon is classified.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/order"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class Superfamily(Field):
    """
    The full scientific name of the superfamily in which the Taxon is classified.

    A taxonomic category subordinate to an order and
    superior to a family. According to ICZN article
    29.2, the suffix `-oidea` is used for a
    superfamily name.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/superfamily"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class Family(Field):
    """
    The full scientific name of the family in which the Taxon is classified.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/family"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class Subfamily(Field):
    """
    The full scientific name of the subfamily in which the Taxon is classified.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/subfamily"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class Tribe(Field):
    """
    The full scientific name of the tribe in which the Taxon is classified.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/tribe"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class Subtribe(Field):
    """
    The full scientific name of the subtribe in which the Taxon is classified.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/subtribe"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class Genus(Field):
    """
    The full scientific name of the genus in which the Taxon is classified.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/genus"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class GenericName(Field):
    """
    The genus part of the Scientific Name without authorship.

    For synonyms the accepted genus and the genus
    part of the name may be different. The term ``Generic Name`` should
    be used together with ``Specific Epithet`` to form a binomial
    and with Infraspecific Epithet to form a trinomial. The
    term Generic Name should only be used for combinations.
    Uninomials of generic rank do not have a Generic Name.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/genericName"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class Subgenus(Field):
    """
    The full scientific name of the subgenus in which the Taxon is classified.

    Values should include the genus to avoid homonym confusion.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/subgenus"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class InfragenericEpithet(Field):
    """
    The infrageneric part of a binomial name at ranks above species but below genus.

    The term ``Infrageneric Epithet`` should be used in
    conjunction with ``Generic Name``, ``Specific Epithet``,
    ``Infraspecific Epithet``, ``Taxon Rank`` and
    ``Scientific Name Authorship`` to represent the
    individual elements of the complete ``Scientific Name``.
    It can be used to indicate the subgenus placement of a
    species, which in zoology is often given in parentheses.
    Can also be used to share infrageneric names such as
    botanical sections (e.g., `Vicia sect. Cracca`).
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/infragenericEpithet"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class SpecificEpithet(Field):
    """
    The name of the first or species epithet of the Scientific Name.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/specificEpithet"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class InfraspecificEpithet(Field):
    """
    The name of the lowest or terminal infraspecific epithet of the Scientific Name, excluding any rank designation.

    In botany, name strings in literature and identifications may
    have multiple infraspecific ranks. According to the International
    Code of Nomenclature for algae, fungi, and plants (Schenzhen
    Code Articles 6.7 & Art. 24.1), valid names only have two epithets,
    with the lowest rank being the ``Infraspecific Epithet``. For
    example: the ``Infraspecific Epithet`` in the string `Indigofera
    charlieriana subsp. sessilis var. scaberrima is scaberrima` and
    the ``Scientific Name`` is `Indigofera charlieriana var. scaberrima
    (Schinz) J.B.Gillett.` Use ``Verbatim Identification`` for the
    full name string used in an Identification.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/infraspecificEpithet"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class CultivarEpithet(Field):
    """
    Part of the name of a cultivar, cultivar group or grex that follows the Scientific Name.

    According to the Rules of the Cultivated Plant Code, a cultivar
    name consists of a botanical name followed by a cultivar epithet.
    The value given as the ``Cultivar Epithet`` should exclude any
    quotes. The term ``Taxon Rank`` should be used to indicate which
    type of cultivated plant name (e.g. cultivar, cultivar group,
    grex) is concerned. This epithet, including any enclosing
    apostrophes or suffix, should be provided in ``Scientific Name``
    as well.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/cultivarEpithet"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class TaxonRank(Field):
    """
    The taxonomic rank of the most specific name in the Scientific Name.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/taxonRank"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class VerbatimTaxonRank(Field):
    """
    The taxonomic rank of the most specific name in the Scientific Name as it appears in the original record.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/verbatimTaxonRank"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class ScientificNameAuthorship(Field):
    """
    The authorship information for the Scientific Name formatted.

    Name formatted according to the conventions of
    the applicable ``Nomenclatural Code``.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/scientificNameAuthorship"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class VernacularName(Field):
    """
    A xml_common or vernacular name.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/vernacularName"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class NomenclaturalCode(Field):
    """
    The nomenclatural code (or codes in the case of an ambiregnal name) under which the Scientific Name is constructed.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/nomenclaturalCode"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class TaxonomicStatus(Field):
    """
    The status of the use of the Scientific Name as a label for a taxon.

    Requires taxonomic opinion to define the
    scope of a Taxon. Rules of priority then
    are used to define the taxonomic status
    of the nomenclature contained in that
    scope, combined with the experts' opinion.
    It must be linked to a specific taxonomic
    reference that defines the concept.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/taxonomicStatus"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class NomenclaturalStatus(Field):
    """
    The status related to the original publication of the name.

    The status related to the original publication
    of the name and its conformance to the relevant
    rules of nomenclature. It is based essentially
    on an algorithm according to the business rules
    of the code. It requires no taxonomic opinion.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/nomenclaturalStatus"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class TaxonRemarks(Field):
    """
    Comments or notes about the taxon or name.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/taxonRemarks"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return



import os.path
import unittest

from lxml import etree as et

from dwca.classes import Identification
from dwca.terms import TaxonID
from test_xml.test_xml import TestXML

PATH = os.path.abspath(os.path.dirname(__file__))


class TestIdentification(TestXML):
    def setUp(self) -> None:
        with open(os.path.join(PATH, os.pardir, "example_data", "meta.xml"), "r", encoding="utf-8") as file:
            content = file.read()
        base_file = et.fromstring(content)
        self.identification_xml = base_file.findall("extension", namespaces=base_file.nsmap)[2]
        self.text = et.tostring(self.identification_xml, pretty_print=True).decode("utf-8")
        return

    def test_name(self):
        taxon = Identification.from_string(self.text)
        self.assertEqual("Identification", taxon.name, "Name parsed incorrectly.")


    def test_parse(self):
        identification = Identification.from_string(self.text)
        self.assertEqual("identification.txt", identification.filename, "Files parse incorrectly")
        self.assertEqual(0, identification.id, "Files parse incorrectly")
        self.assertEqual("UTF-8", identification.__encoding__, "Encoding parse incorrectly")
        self.assertEqual("\n", identification.__lines_end__, "Lines end parse incorrectly")
        self.assertEqual("\t", identification.__fields_end__, "Field end parse incorrectly")
        self.assertEqual("", identification.__fields_enclosed__, "Field enclosed parse incorrectly")
        self.assertEqual(1, identification.__ignore_header_lines__, "Ignore header parse incorrectly")
        self.assertEqual(Identification.URI, identification.uri, "Row type parse incorrectly")
        self.assertEqual(13, len(identification.__fields__), "Fields parse incorrectly")
        self.assertEqualTree(self.identification_xml, identification.to_element(), "Error on element conversion")

    def test_set_core_field(self):
        identification = Identification.from_string(self.text)
        self.assertEqual("", identification.fields[0], "Field from nowhere.")
        identification.set_core_field(TaxonID(3))
        self.assertEqual(TaxonID(3).uri, identification.fields[0], "Field not set.")
        self.assertEqual(0, identification.__fields__[0].index, "Field not changed on set.")

    def test_none(self):
        self.assertIsNone(Identification.parse(None, {}), "Object parsed from nothing")

    def test_missing_field(self):
        self.assertRaises(
            AssertionError,
            Identification.from_string,
            """
<core>
    <files>
      <location>missing.txt</location>
    </files>
    <id index="0"/>
    <field index="0" term=""/>
    <field index="1" term=""/>
    <field index="3" term=""/>
</core>
            """
        )

    def test_field_twice(self):
        self.assertRaises(
            AssertionError,
            Identification.from_string,
            """
<core>
    <files>
      <location>missing.txt</location>
    </files>
    <id index="0"/>
    <field index="0" term=""/>
    <field index="1" term=""/>
    <field index="3" term=""/>
    <field index="2" term=""/>
    <field index="3" term=""/>
    <field index="4" term=""/>
</core>
            """
        )

    def test_parse_invalid(self):
        self.assertRaises(
            AssertionError,
            Identification.from_string,
            '<extension />'
        )
        self.assertRaises(
            KeyError,
            Identification.from_string,
            """
<core-id>
    <field index="0" term=""/>
</core-id>
            """
        )

    def test_sql_error(self):
        identification = Identification.from_string(self.text)
        with self.assertRaisesRegex(RuntimeError, "Extension"):
            _ = identification.sql_table
        identification.set_primary_key("Taxon")
        with self.assertRaisesRegex(RuntimeError, "Extension"):
            _ = identification.sql_table

    def test_set_primary_key(self):
        identification = Identification.from_string(self.text)
        self.assertIsNone(identification.__primary_key__, "Primary key from nowhere.")
        identification.set_primary_key("Taxon")
        self.assertIsNotNone(identification.__primary_key__, "Primary key not set.")
        self.assertEqual("Taxon", identification.__primary_key__, "Incorrect Primary key.")

    def test_sql_table(self):
        identification = Identification.from_string(self.text)
        identification.set_primary_key("Taxon")
        identification.set_core_field(TaxonID(3))
        self.assertEqual(
            self.normalize_sql("""
            CREATE TABLE "Identification" (
                "taxonID" VARCHAR,
                "identificationID" VARCHAR,
                "verbatimIdentification" VARCHAR,
                "identificationQualifier" VARCHAR,
                "typeStatus" VARCHAR,
                "identifiedBy" VARCHAR,
                "identifiedByID" VARCHAR,
                "dateIdentified" VARCHAR,
                "identificationReferences" VARCHAR,
                "identificationVerificationStatus" VARCHAR,
                "identificationRemarks" VARCHAR,
                "datasetID" VARCHAR,
                "subject" VARCHAR,
                PRIMARY KEY ("taxonID"),
                FOREIGN KEY ("taxonID") REFERENCES "Taxon"
            );
            """),
            self.normalize_sql(identification.sql_table),
            "Wrong SQL table."
        )


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

from enum import Enum
from typing import List, Dict, Any


class EMLVersion(Enum):
    """
    Version of the specification of the EML used.
    """
    VERSION_2_1_1 = (2, 1, 1)
    VERSION_2_2_0 = (2, 2, 0)
    LATEST = VERSION_2_2_0

    def schema_location(self) -> List[str]:
        """
        The schema location of the given version.

        Returns
        -------
        List[str]
            A list of url where the schema of the XML version is located.

        Raises
        ------
        NotImplementedError
            Version not implemented or not existing.
        """
        if self.value == (2, 1, 1):
            return [
                "eml://ecoinformatics.org/eml-2.1.1",
                "http://rs.gbif.org/schema/eml-gbif-profile/1.0.1/eml.xsd"
            ]
        else:  # self.value == (2, 2, 0):
            return [
                "https://eml.ecoinformatics.org/eml-2.2.0",
                "xsd/eml.xsd"
            ]

    def get_namespace(self) -> Dict[Any, str]:
        """
        Get namespace of the given version.

        Returns
        -------
        Dict[Any, str]
            Namespace as a dictionary of value:uri.
        """
        if self.value == (2, 1, 1):
            nmap = {
                "eml": "eml://ecoinformatics.org/eml-2.1.1",
                "md": "eml://ecoinformatics.org/methods-2.1.1",
                "proj": "eml://ecoinformatics.org/project-2.1.1",
                "d": "eml://ecoinformatics.org/dataset-2.1.1",
                "res": "eml://ecoinformatics.org/resource-2.1.1",
                "dc": "http://purl.org/dc/terms/",
            }
        else:  # self.value == (2, 2, 0):
            nmap = {
                "eml": "https://eml.ecoinformatics.org/eml-2.2.0",
                "stmml": "http://www.xml-cml.org/schema/stmml-1.1"
            }
        nmap["xsi"] = "http://www.w3.org/2001/XMLSchema-instance"
        return nmap

    @classmethod
    def get_version(cls, schema_location: str) -> EMLVersion:
        """
        Get version of EML parsing the schema location provided.

        Parameters
        ----------
        schema_location : str
            Schema location text.

        Returns
        -------
        EMLVersion
            Version of the EML.
        """
        if schema_location is None:
            return EMLVersion.LATEST
        schemas = schema_location.split(" ")
        for schema in schemas:
            if "ecoinformatics.org/eml-" in schema:
                version_str = schema.partition("ecoinformatics.org/eml-")[-1]
                return EMLVersion(tuple([int(v) for v in version_str.split(".")]))
        raise NotImplementedError(f"Version not found in: '{schema_location}'")



import re
import unittest
from copy import deepcopy

from lxml import etree as et


class TestXML(unittest.TestCase):
    DEFAULT_TAGS = {}

    def assertEqualTree(self, expected: et.ElementTree, actual: et.ElementTree, msg: str) -> None:
        try:
            self.assertEqual(expected.tag, actual.tag, "Not same tag")
            expected_text = expected.text.strip() if expected.text is not None else ""
            actual_text = actual.text.strip() if actual.text is not None else ""
            self.assertEqual(expected_text, actual_text, "Not same text")
            # Add default values:
            exp_attributes = deepcopy(expected.attrib)
            actual_attributes = deepcopy(actual.attrib)
            for key, value in self.DEFAULT_TAGS.items():
                if key not in exp_attributes:
                    exp_attributes[key] = value
                if key not in actual_attributes:
                    actual_attributes[key] = value
            self.assertEqual(exp_attributes, actual_attributes, "Not same attributes")
            expected_tags = set([child.tag for child in expected])
            actual_tags = set([child.tag for child in actual])
            self.assertCountEqual(expected_tags, actual_tags, "Different tags")
            for tag in expected_tags:
                for expected_child, actual_child in zip(expected.findall(tag), actual.findall(tag)):
                    self.assertEqualTree(expected_child, actual_child, msg=msg)
        except AssertionError:
            raise AssertionError(f"{et.tostring(expected).decode()} != {et.tostring(actual).decode()} : {msg}")

    @staticmethod
    def normalize_sql(sql: str) -> str:
        """
        Normalize an SQL statement string from comparison purpose.

        Parameters
        ----------
        sql : str
            An SQL statement string.

        Returns
        -------
        str
            The same statement with normalize space.
        """
        sql = re.sub(r'\s+', ' ', sql)
        return sql.strip()



import datetime as dt
from typing import Union, List

from datetime_interval import Interval

from dwca.terms import Field


class EventID(Field):
    """
    An identifier for the set of information associated with an Event (something that occurs at a place and time).
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/eventID"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class ParentEventID(Field):
    """
    An identifier for the broader Event that groups this and potentially other Events.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/parentEventID"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class EventType(Field):
    """
    The nature of the Event.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/eventType"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class FieldNumber(Field):
    """
    An identifier given to the Event in the field. Often serves as a link between field notes and the Event.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/fieldNumber"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class EventDate(Field):
    """
    The date-time or interval during which an Event occurred.

    For occurrences, this is the date-time when the Event was recorded.
    Not suitable for a time in a geological context.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/eventDate"
    TYPE = Union[dt.datetime, Interval]

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class EventTime(Field):
    """
    The time or interval during which a Event occurred.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/eventTime"
    TYPE = Union[dt.time, Interval]

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class StartDayOfYear(Field):
    """
    The earliest integer day of the year on which the Event occurred.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/startDayOfYear"
    TYPE = int

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class EndDayOfYear(Field):
    """
    The latest integer day of the year on which the Event occurred.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/endDayOfYear"
    TYPE = int

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class DWCYear(Field):
    """
    The four-digit year in which the Event occurred, according to the Common Era Calendar.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/year"
    TYPE = int

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class DWCMonth(Field):
    """
    The integer month in which the Event occurred.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/month"
    TYPE = int

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class DWCDay(Field):
    """
    The integer day of the month on which the Event occurred.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/day"
    TYPE = int

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class VerbatimEventDate(Field):
    """
    The verbatim original representation of the date and time information for an Event.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/verbatimEventDate"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class Habitat(Field):
    """
    A category or description of the habitat in which the Event occurred.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/habitat"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class SamplingProtocol(Field):
    """
    The names of, references to, or descriptions of the methods or protocols used during a Event.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/samplingProtocol"
    TYPE = List[str]

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class SampleSizeValue(Field):
    """
    A measurement of the size (time duration, length, area, or volume) of a sample in a sampling Event.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/sampleSizeValue"
    TYPE = float

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class SampleSizeUnit(Field):
    """
    The unit of measurement of the size (time duration, length, area, or volume) of a sample in a sampling Event.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/sampleSizeUnit"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class SamplingEffort(Field):
    """
    The amount of effort expended during an Event.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/samplingEffort"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class FieldNotes(Field):
    """
    Notes taken in the field about the dwc:Event.

    One of:
    a) An indicator of the existence of the notes.
    b) A reference to (publication, URI) the notes.
    c) The text of the notes.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/fieldNotes"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class EventRemarks(Field):
    """
    Comments or notes about the Event.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/eventRemarks"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return



from __future__ import annotations
from typing import Dict

from lxml import etree as et

from xml_common import XMLObject


class EMLLicense(XMLObject):
    """
    Information identifying a well-known license for the metadata and data.

    Parameters
    ----------
    name : str
        The official name of the license.
    url : str, optional
        The persistent URL for the license.
    identifier : str, optional
        License Identifier.
    """
    PRINCIPAL_TAG = "licensed"

    def __init__(self, name: str, url: str = None, identifier: str = None) -> None:
        super().__init__()
        self.__name__ = name
        self.__url__ = url
        self.__id__ = identifier
        return

    @property
    def name(self) -> str:
        """str: The official name of the license."""
        return self.__name__

    @property
    def url(self) -> str:
        """str: The persistent URL for the license."""
        return self.__url__

    @property
    def id(self) -> str:
        """str: License Identifier."""
        return self.__id__

    @classmethod
    def parse(cls, element: et.Element, nmap: Dict) -> EMLLicense | None:
        """
        Generate an EMLLicense object from a given XML element.

        Parameters
        ----------
        element : lxml.etree.Element
            XML element to parse
        nmap : Dict
            Namespace

        Returns
        -------
        EMLLicense
            Parsed object
        """
        if element is None:
            return None
        name_elem = element.find("licenseName", nmap)
        name = name_elem.text if name_elem.text is not None else ""
        url_elem = element.find("url", nmap)
        if url_elem is not None:
            url = url_elem.text if url_elem.text is not None else ""
        else:
            url = None
        id_elem = element.find("identifier", nmap)
        if id_elem is not None:
            identifier = id_elem.text if id_elem.text is not None else ""
        else:
            identifier = None
        licence = EMLLicense(name, url=url, identifier=identifier)
        licence.__namespace__ = nmap
        return licence

    def to_element(self) -> et.Element:
        """
        Generate an XML element with EMLLicense instance information.

        Returns
        -------
        lxml.etree.Element
            XML element object
        """
        license_elem = super().to_element()
        license_name = self.object_to_element("licenseName")
        license_name.text = self.name
        license_elem.append(license_name)
        if self.url is not None:
            url = self.object_to_element("url")
            url.text = self.url
            license_elem.append(url)
        if self.id is not None:
            identifier = self.object_to_element("identifier")
            identifier.text = self.id
            license_elem.append(identifier)
        return license_elem



import unittest
from lxml import etree as et

from xml_common.utils import Language
from eml.types import EMLTextType, EMLSection
from test_xml.test_xml import TestXML


class TestAbstract(TestXML):
    PARAGRAPH = ("Lorem ipsum dolor sit amet, consectetur "
                 "adipiscing elit. Sed euismod urna sit amet "
                 "pulvinar porttitor. Ut aliquet lectus sit amet.")
    ENG_PARAGRAPH = ("The company's marketing team will be "
                     "added to the company's budget. At a "
                     "price tag, it's now a real estate company, "
                     "at times.")
    ESP_PARAGRAPH = ("El equipo de marketing de la empresa se "
                     "agregará al presupuesto de la empresa. Por "
                     "un precio, ahora es una empresa de bienes "
                     "raíces, a veces.")
    MARKDOWN = ("*Lorem ipsum* dolor sit amet, _consectetur_ "
                "adipiscing elit. Sed euismod urna sit amet "
                "pulvinar porttitor. Ut aliquet lectus sit amet. "
                "Generated using [Lorem ipsum generator](https://www.lipsum.com/)")
    DEFAULT_TAGS = {
        "{http://www.w3.org/XML/1998/namespace}lang": "eng",
    }

    def test_parse_simple(self):
        text_xml = f"""
<abstract>
    <para>{self.PARAGRAPH}</para>
</abstract>
        """
        abstract = EMLTextType.from_string(text_xml)
        self.assertEqual(0, len(abstract.sections), "Section generated from nowhere")
        self.assertEqual(self.PARAGRAPH, abstract.paragraphs[0], "Paragraph parsed incorrectly")
        self.assertEqual(0, len(abstract.markdowns), "Markdown generated from nowhere")
        self.assertEqual(Language.ENG, abstract.language, "Language parsed incorrectly")
        self.assertRaises(
            RuntimeError,
            abstract.to_element
        )
        abstract.set_tag("abstract")
        self.assertEqualTree(et.fromstring(text_xml), abstract.to_element(), "To element incorrect")

    def test_parse_empty(self):
        text_xml = "<abstract></abstract>"
        abstract = EMLTextType.from_string(text_xml)
        self.assertEqual(0, len(abstract.sections), "Section generated from nowhere")
        self.assertEqual(0, len(abstract.paragraphs), "Paragraph generated from nowhere")
        self.assertEqual(0, len(abstract.markdowns), "Markdown generated from nowhere")
        self.assertRaises(
            RuntimeError,
            abstract.to_element
        )
        abstract.set_tag("abstract")
        self.assertEqualTree(et.fromstring(text_xml), abstract.to_element(), "To element incorrect")

    def test_parse_complete(self):
        text_xml = f"""
<abstract>
    <section>
        <para>A small section</para>
    </section>
    <para>{self.PARAGRAPH}</para>
    <markdown>{self.MARKDOWN}</markdown>
</abstract>
        """
        abstract = EMLTextType.from_string(text_xml)
        self.assertEqual(
            EMLSection(paragraph="A small section"),
            abstract.sections[0],
            "Section parsed incorrectly"
        )
        self.assertEqual(self.PARAGRAPH, abstract.paragraphs[0], "Paragraph parsed incorrectly")
        self.assertEqual(self.MARKDOWN, abstract.markdowns[0], "Markdown parsed incorrectly")
        self.assertRaises(
            RuntimeError,
            abstract.to_element
        )
        abstract.set_tag("abstract")
        self.assertEqualTree(et.fromstring(text_xml), abstract.to_element(), "To element incorrect")

    def test_parse_none(self):
        self.assertIsNone(EMLTextType.parse(None, {}), "Error on parse None")

    def test_multilanguage(self):
        text_xml = f"""
<abstract>
    <para xml:lang="eng">{self.ENG_PARAGRAPH}</para>
    <para xml:lang="esp">{self.ESP_PARAGRAPH}</para>
    <markdown>{self.MARKDOWN}</markdown>
</abstract>
        """
        abstract = EMLTextType.from_string(text_xml)
        self.assertEqual(0, len(abstract.sections), "Section generated from nowhere")
        self.assertEqual(self.ENG_PARAGRAPH, abstract.paragraphs[0], "English Paragraph parsed incorrectly")
        self.assertEqual(Language.ENG, abstract.paragraphs[0].language, "English Paragraph wrong language")
        self.assertEqual(self.ESP_PARAGRAPH, abstract.paragraphs[1], "Spanish Paragraph parsed incorrectly")
        self.assertEqual(Language.ESP, abstract.paragraphs[1].language, "Spanish Paragraph wrong language")
        self.assertRaises(
            RuntimeError,
            abstract.to_element
        )
        abstract.set_tag("abstract")
        self.assertEqualTree(et.fromstring(text_xml), abstract.to_element(), "To element incorrect")


if __name__ == '__main__':
    unittest.main()



from typing import List

from dwca.classes import DataFile, DataFileType
from dwca.terms import OrganismID, OrganismName, OrganismScope, AssociatedOrganisms, PreviousIdentifications, \
    OrganismRemarks, Field


class Organism(DataFile):
    """
    A particular organism or defined group of organisms considered to be taxonomically homogeneous.

    Parameters
    ----------
    _id : int
        Unique identifier for the core entity.
    files : str
        File location, in the archive, this is inside the `zip` file.
    fields : List[Field]
        A list of the Field (columns) in the Core data entity.
    data_file_type: DataFileType
        The Data File Type in the Darwin Core Archive.
    encoding : str, optional
        Encoding of the file location (`files` parameter), default is "utf-8".
    lines_terminated_by : str, optional
        Delimiter of lines on the file, default `"\\\\n"`.
    fields_terminated_by : str, optional
        Delimiter of the file (cells) on the file, default `","`.
    fields_enclosed_by : str, optional
        Specifies the character used to enclose (mark the start and end of) each field, default empty `""`.
    ignore_header_lines : int, optional
        Ignore headers at the start of document, can be one line or a list of them, default 0 (first line).
    """
    URI = DataFile.URI + "Organism"
    __field_class__ = DataFile.__field_class__ + [
        OrganismID, OrganismName, OrganismScope,
        AssociatedOrganisms, PreviousIdentifications,
        OrganismRemarks,
    ]

    def __init__(
            self, _id: int, files: str,
            fields: List[Field],
            data_file_type: DataFileType = DataFileType.CORE,
            encoding: str = "utf-8",
            lines_terminated_by: str = "\n",
            fields_terminated_by: str = ",",
            fields_enclosed_by: str = "",
            ignore_header_lines: int = 0,
    ) -> None:
        super().__init__(
            _id, files, fields, data_file_type, encoding,
            lines_terminated_by, fields_terminated_by,
            fields_enclosed_by, ignore_header_lines,
        )
        return



from __future__ import annotations
from typing import Dict, Tuple, Any

from lxml import etree as et

from eml.types import EMLObject, _NoTagObject, Scope, ExtensionURI


class SemanticAnnotation(EMLObject, _NoTagObject):
    """
    A precisely-defined semantic statement about this resource.

    Parameters
    ----------
    property_uri : Tuple[str, str]
        The persistent URI used to identify a property from a vocabulary.
    value_uri : Tuple[str, str]
        The persistent URI used to identify a value from a vocabulary.
    _id : str, optional
        Unique identifier within the scope.
    scope : Scope, default DOCUMENT
        The scope of the identifier.
    system : str, optional
        The data management system within which an identifier is in scope and therefore unique.
    """
    def __init__(
            self,
            property_uri: Tuple[str, str],
            value_uri: Tuple[str, str],
            _id: str = None,
            scope: Scope = Scope.DOCUMENT,
            system: str = None,
    ) -> None:
        super().__init__(_id, scope, system)
        self.__property__ = ExtensionURI(property_uri[0], property_uri[1])
        self.__value__ = ExtensionURI(value_uri[0], value_uri[1])
        return

    @property
    def property_uri(self) -> ExtensionURI:
        """ExtensionURI: The persistent URI used to identify a property from a vocabulary."""
        return self.__property__

    @property
    def value_uri(self) -> ExtensionURI:
        """ExtensionURI: The persistent URI used to identify a value from a vocabulary."""
        return self.__value__

    @classmethod
    def get_referrer(cls, element: et.Element, nmap: Dict) -> SemanticAnnotation:
        """
        Semantic Annotation doesn't have a reference. This method raise an exception.

        Parameters
        ----------
        element : lxml.etree.Element
            XML element to parse with references object.
        nmap : Dict
            Namespace.

        Returns
        -------
        SemanticAnnotation
            Object parsed that would reference another.

        Raises
        ------
        ValueError
            Semantic Annotation cannot reference another annotation.
        """
        raise ValueError("Semantic Annotation cannot reference another annotation")

    @classmethod
    def get_no_referrer(cls, element: et.Element, nmap: Dict) -> SemanticAnnotation:
        """
        Generate a Semantic Annotation from an XML element instance.

        Parameters
        ----------
        element : lxml.etree.Element
            XML element to parse.
        nmap : Dict
            Namespace.

        Returns
        -------
        SemanticAnnotation
            Object parsed.
        """
        property_elem = element.find("propertyURI", nmap)
        value_elem = element.find("valueURI", nmap)
        return SemanticAnnotation(
            (property_elem.text, property_elem.get("label")),
            (value_elem.text, value_elem.get("label")),
            _id=element.get("id", None),
            scope=cls.get_scope(element),
            system=element.get("system", None)
        )

    def to_element(self) -> et.Element:
        """
        Generate an XML element instance.

        Returns
        -------
        lxml.etree.Element
            XML element instance.
        """
        element = _NoTagObject.to_element(self)
        element = self._to_element_(element)
        self.property_uri.set_tag("propertyURI")
        element.append(self.property_uri.to_element())
        self.value_uri.set_tag("valueURI")
        element.append(self.value_uri.to_element())
        return element

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, SemanticAnnotation):
            return (super().__eq__(other) and
                    self.property_uri == other.property_uri and
                    self.value_uri == other.value_uri)
        else:
            return False



from __future__ import annotations
from typing import Dict, List

from lxml import etree as et

from xml_common import XMLObject


class EMLOffline(XMLObject):
    """
    Offline distribution when data are available offline.

    Parameters
    ----------
    medium_name : str
        Name of the medium that for this resource distribution.
    medium_density : str, optional
        The density of the digital medium if this is relevant.
    medium_density_units : str, optional
        A numerical density's units.
    medium_volume : str, optional
        Total volume of the storage medium.
    medium_format : List[str], optional
        Format of the medium on which the resource is shipped.
    medium_note : str, optional
        Note about the media.
    """
    PRINCIPAL_TAG = 'offline'
    """str: Principle tag `offline`."""

    def __init__(
            self, medium_name: str,
            medium_density: str = None,
            medium_density_units: str = None,
            medium_volume: str = None,
            medium_format: List[str] = None,
            medium_note: str = None,
    ) -> None:
        super().__init__()
        self.__name__ = medium_name
        self.__density__ = medium_density
        self.__density_units__ = medium_density_units
        self.__volume__ = medium_volume
        self.__format__ = list()
        if medium_format is not None:
            self.__format__.extend(medium_format)
        self.__note__ = medium_note
        return

    @property
    def medium_name(self) -> str:
        """str: Name of the medium that for this resource distribution."""
        return self.__name__

    @property
    def medium_density(self) -> str:
        """str : Density of the digital medium if this is relevant."""
        return self.__density__

    @property
    def medium_density_units(self) -> str:
        """str : Numerical density's units."""
        return self.__density_units__

    @property
    def medium_volume(self) -> str:
        """str: Total volume of the storage medium."""
        return self.__volume__

    @property
    def medium_format(self) -> List[str]:
        """List[str]: Format of the medium on which this resource is shipped."""
        return self.__format__

    @property
    def medium_note(self) -> str:
        """str: Note about the media."""
        return self.__note__

    @classmethod
    def parse(cls, element: et.Element, nmap: Dict) -> EMLOffline | None:
        """
        Generate an EMLOffline object from an XML element.

        Parameters
        ----------
        element : lxml.etree.Element
            An XML element.
        nmap : Dict
            Namespace.

        Returns
        -------
        EMLOffline
            Instance with the offline media distribution.
        """
        if element is None:
            return None
        name_elem = element.find("mediumName", nmap)
        name = name_elem.text if name_elem.text is not None else ""
        density_elem = element.find("mediumDensity", nmap)
        if density_elem is not None:
            density = density_elem.text if density_elem.text is not None else ""
        else:
            density = None
        units_elem = element.find("mediumDensityUnits", nmap)
        if units_elem is not None:
            units = units_elem.text if units_elem.text is not None else ""
        else:
            units = None
        volume_elem = element.find("mediumVolume", nmap)
        if volume_elem is not None:
            volume = volume_elem.text if volume_elem.text is not None else ""
        else:
            volume = None
        medium_format = list()
        for format_elem in element.findall("mediumFormat", nmap):
            medium_format.append(format_elem.text if format_elem.text is not None else "")
        note_elem = element.find("mediumNote", nmap)
        if note_elem is not None:
            note = note_elem.text if note_elem.text is not None else ""
        else:
            note = None
        offline = EMLOffline(
            medium_name=name,
            medium_density=density,
            medium_density_units=units,
            medium_volume=volume,
            medium_format=medium_format,
            medium_note=note,
        )
        offline.__namespace__ = nmap
        return offline

    def to_element(self) -> et.Element:
        """
        Generate an XML element with the information of the EMLOffline instance.

        Returns
        -------
        lxml.etree.Element
            XML element with the information of the EMLOffline.
        """
        offline_elem = super().to_element()
        name = self.object_to_element("mediumName")
        name.text = self.medium_name
        offline_elem.append(name)
        if self.medium_density is not None:
            density_elem = self.object_to_element("mediumDensity")
            density_elem.text = self.medium_density
            offline_elem.append(density_elem)
        if self.medium_density_units is not None:
            units_elem = self.object_to_element("mediumDensityUnits")
            units_elem.text = self.medium_density_units
            offline_elem.append(units_elem)
        if self.medium_volume is not None:
            volume_elem = self.object_to_element("mediumVolume")
            volume_elem.text = self.medium_volume
            offline_elem.append(volume_elem)
        for medium_format in self.medium_format:
            format_elem = self.object_to_element("mediumFormat")
            format_elem.text = medium_format
            offline_elem.append(format_elem)
        if self.medium_note is not None:
            note_elem = self.object_to_element("mediumNote")
            note_elem.text = self.medium_note
            offline_elem.append(note_elem)
        return offline_elem



from dwca.terms.field import Field
from dwca.terms.outside_term import OutsideTerm
from dwca.terms.record_level import DWCType, DWCModified, DWCLanguage, \
    DWCLicense, DWCRightsHolder, DWCAccessRights, DWCBibliographicCitation, \
    DWCReferences, DWCInstitution, DWCCollection, DWCDataset, \
    DWCInstitutionCode, DWCCollectionCode, DWCDatasetName, \
    DWCOwnerInstitutionCode, DWCBasisOfRecord, DWCInformationWithheld, \
    DWCDataGeneralizations, DWCDynamicProperties
from dwca.terms.occurrence import OccurrenceID, CatalogNumber, RecordNumber, \
    RecordedBy, RecordedByID, IndividualCount, OrganismQuantity, \
    OrganismQuantityType, OccurrenceSex, LifeStage, ReproductiveCondition, \
    Caste, Behavior, Vitality, DWCEstablishmentMeans, DWCDegreeOfEstablishment, \
    Pathway, GeoreferenceVerificationStatus, OccurrenceStatus, AssociatedMedia, \
    AssociatedOccurrences, AssociatedReferences, AssociatedTaxa, \
    OtherCatalogNumbers, OccurrenceRemarks
from dwca.terms.organism import OrganismID, OrganismName, OrganismScope, \
    AssociatedOrganisms, PreviousIdentifications, OrganismRemarks
from dwca.terms.material_entity import MaterialEntityID, Preparations, \
    Disposition, VerbatimLabel, AssociatedSequences, MaterialEntityRemarks
from dwca.terms.material_sample import MaterialSampleID
from dwca.terms.event import EventID, ParentEventID, EventType, FieldNumber, \
    EventDate, EventTime, StartDayOfYear, EndDayOfYear, DWCYear, DWCMonth, \
    DWCDay, VerbatimEventDate, Habitat, SamplingProtocol, SampleSizeValue, \
    SampleSizeUnit, SamplingEffort, FieldNotes, EventRemarks
from dwca.terms.location import LocationID, HigherGeographyID, HigherGeography, \
    Continent, WaterBody, IslandGroup, Island, Country, CountryCode, StateProvince, \
    County, Municipality, DWCLocalityTerm, VerbatimLocality, MinimumElevationInMeters, \
    MaximumElevationInMeters, VerbatimElevation, VerticalDatum, MinimumDepthInMeters, \
    MaximumDepthInMeters, VerbatimDepth, MinimumDistanceAboveSurfaceInMeters, \
    MaximumDistanceAboveSurfaceInMeters, LocationAccordingTo, LocationRemarks, \
    DecimalLatitude, DecimalLongitude, GeodeticDatum, CoordinateUncertaintyInMeters, \
    CoordinatePrecision, PointRadiusSpatialFit, VerbatimCoordinates, VerbatimLatitude, \
    VerbatimLongitude, VerbatimCoordinateSystem, VerbatimSRS, FootprintWKT, \
    FootprintSRS, FootprintSpatialFit, GeoreferencedBy, GeoreferencedDate, \
    GeoreferenceProtocol, GeoreferenceSources, GeoreferenceRemarks
from dwca.terms.geological_context import GeologicalContextID, EarliestEonOrLowestEonothem, \
    LatestEonOrHighestEonothem, EarliestEraOrLowestErathem, LatestEraOrHighestErathem, \
    EarliestPeriodOrLowestSystem, LatestPeriodOrHighestSystem, EarliestEpochOrLowestSeries, \
    LatestEpochOrHighestSeries, EarliestAgeOrLowestStage, LatestAgeOrHighestStage, \
    LowestBiostratigraphicZone, HighestBiostratigraphicZone, LithostratigraphicTerms, \
    LithostratigraphicGroup, LithostratigraphicFormation, LithostratigraphicMember, \
    LithostratigraphicBed
from dwca.terms.identification import IdentificationID, VerbatimIdentification, \
    IdentificationQualifier, TypeStatus, IdentifiedBy, IdentifiedByID, \
    DateIdentified, IdentificationReferences, IdentificationVerificationStatus, \
    IdentificationRemarks
from dwca.terms.taxon import TaxonID, ScientificNameID, AcceptedNameUsageID, ParentNameUsageID, \
    OriginalNameUsageID, NameAccordingToID, NamePublishedInID, TaxonConceptID, \
    ScientificName, AcceptedNameUsage, ParentNameUsage, OriginalNameUsage, NameAccordingTo, \
    NamePublishedIn, NamePublishedInYear, HigherClassification, Kingdom, Phylum, DWCClass, Order, \
    Superfamily, Family, Subfamily, Tribe, Subtribe, Genus, GenericName, Subgenus, \
    InfragenericEpithet, SpecificEpithet, InfraspecificEpithet, CultivarEpithet, \
    TaxonRank, VerbatimTaxonRank, ScientificNameAuthorship, VernacularName, NomenclaturalCode, \
    TaxonomicStatus, NomenclaturalStatus, TaxonRemarks
from dwca.terms.resource_relationship import ResourceRelationshipID, ResourceID, \
    RelationshipOfResourceID, RelatedResourceID, RelationshipOfResource, RelationshipAccordingTo, \
    RelationshipEstablishedDate, RelationshipRemarks
from dwca.terms.measurement_or_fact import MeasurementID, ParentMeasurementID, \
    MeasurementType, MeasurementValue, MeasurementAccuracy, MeasurementUnit, \
    MeasurementDeterminedBy, MeasurementDeterminedDate, MeasurementMethod, \
    MeasurementRemarks
from dwca.terms.chronometric_age import ChronometricAgeID, VerbatimChronometricAge, \
    ChronometricAgeProtocol, UncalibratedChronometricAge, ChronometricAgeConversionProtocol, \
    EarliestChronometricAge, EarliestChronometricAgeReferenceSystem, LatestChronometricAge, \
    LatestChronometricAgeReferenceSystem, ChronometricAgeUncertaintyInYears, \
    ChronometricAgeUncertaintyMethod, MaterialDated, MaterialDatedID, MaterialDatedRelationship, \
    ChronometricAgeDeterminedBy, ChronometricAgeDeterminedDate, ChronometricAgeReferences, \
    ChronometricAgeRemarks



import os.path
import unittest
from unittest.mock import patch

from test_dwca_classes.test_taxon_common import TestTaxonCommon

PATH = os.path.abspath(os.path.dirname(__file__))

orig_import = __import__


def import_mock(name, globals=None, locals=None, fromlist=(), level=0):
    if name == 'pandas':
        raise ImportError(f"No module named '{name}'")
    return orig_import(name, globals, locals, fromlist, level)


def import_mock_fuzzy(name, globals=None, locals=None, fromlist=(), level=0):
    if name == 'rapidfuzz':
        raise ImportError(f"No module named '{name}'")
    return orig_import(name, globals, locals, fromlist, level)


class TestTaxonNoPandas(TestTaxonCommon):
    @patch('builtins.__import__', side_effect=import_mock)
    def test_add_field(self, mock_import):
        super().__test_add_field__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_add_field_incorrect_index(self, mock_import):
        super().__test_add_field_incorrect_index__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_parse(self, mock_import):
        super().__test_parse__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_filter_kingdom_exception(self, mock_import):
        super().__test_filter_kingdom_exception__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_get_parents(self, mock_import):
        super().__test_get_parents__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_get_incorrect_parent(self, mock_import):
        self.__test_get_incorrect_parent__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_get_synonyms(self, mock_import):
        self.__test_get_synonyms__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_get_synonyms_names(self, mock_import):
        super().__test_get_synonyms_names__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_get_incorrect_synonyms(self, mock_import):
        self.__test_get_incorrect_synonyms__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_filter_kingdom(self, mock_import):
        self.read_pandas()
        length = len(self.taxon)
        self.assertEqual(length, len(self.taxon), "Original length of dataframe.")
        self.taxon.filter_by_kingdom(["Ydhxpqku", "Zcmmvusczo"])
        self.assertEqual(length, len(self.taxon), "Filter by all kingdoms.")
        self.taxon.filter_by_kingdom(["Ydhxpqku"])
        self.assertEqual(159490, len(self.taxon), "Filter by first kingdom.")

    @patch('builtins.__import__', side_effect=import_mock)
    def test_filter_phylum(self, mock_import):
        self.read_pandas()
        length = len(self.taxon)
        self.assertEqual(length, len(self.taxon), "Original length of dataframe.")
        self.taxon.filter_by_phylum(["Ajiwcqftc", "Xpqnzcij", "Gdhevlcsj", "Xtgktcts"])
        self.assertEqual(length, len(self.taxon), "Filter by all phylum.")
        self.taxon.filter_by_phylum(["Ajiwcqftc"])
        self.assertEqual(
            79772,
            len(self.taxon),
            "Incorrect filter by Ajiwcqftc."
        )

    @patch('builtins.__import__', side_effect=import_mock)
    def test_filter_phylum_no_threshold(self, mock_import):
        self.read_pandas()
        length = len(self.taxon)
        self.assertEqual(length, len(self.taxon), "Original length of dataframe.")
        # Modified phylum 0
        # Change 2 letter
        self.taxon.filter_by_phylum(["AjiYqftc", "Xpqnzcij", "Gdhevlcsj", "Xtgktcts"], fuzzy_threshold=100)
        self.assertGreater(length, len(self.taxon), "Filter by all phylum with no threshold.")

    @patch('builtins.__import__', side_effect=import_mock)
    def test_filter_phylum_threshold(self, mock_import):
        self.read_pandas()
        length = len(self.taxon)
        self.assertEqual(length, len(self.taxon), "Original length of dataframe.")
        # Modified phylum 0
        # Change 2 letter
        self.taxon.filter_by_phylum(["AjiYqftc", "Xpqnzcij", "Gdhevlcsj", "Xtgktcts"], fuzzy_threshold=50)
        self.assertEqual(length, len(self.taxon), "Filter by all phylum with enough threshold.")
        self.taxon.filter_by_phylum(["AjiYqftc", "Xpqnzcij", "Gdhevlcsj", "Xtgktcts"], fuzzy_threshold=90)
        self.assertGreater(length, len(self.taxon), "Filter by all phylum with not enough threshold.")

    @patch('builtins.__import__', side_effect=import_mock)
    def test_filter_class(self, mock_import):
        self.read_pandas()
        length = len(self.taxon)
        self.assertEqual(length, len(self.taxon), "Original length of dataframe.")
        self.taxon.filter_by_class([
            "Rwqpdxtmzf", "Wturcymgs", "Yichyzkbff",
            "Mrtuudeumnwn", "Yoadtafayw", "Vicipbwnr",
            "Qipttfiyorbt", "Pvopxyxw",  "Uzskxjpkntg"
        ])
        self.assertEqual(length, len(self.taxon), "Filter by all classes.")
        self.taxon.filter_by_class(["Yichyzkbff"])
        self.assertEqual(
            1 + 39941 + 1 + 1,
            len(self.taxon),
            "Incorrect filter by Yichyzkbff."
        )

    @patch('builtins.__import__', side_effect=import_mock)
    def test_filter_order(self, mock_import):
        self.read_pandas()
        length = len(self.taxon)
        self.assertEqual(length, len(self.taxon), "Original length of dataframe.")
        self.taxon.filter_by_order([
            "Emicapsmvx", "Qyztmnpzz", "Udfewrqmk", "Ffknrbypsk", "Tqmvrbmttmo", "Xhngcistiqri", "Iwxoydtagzew",
            "Civppxeuqtes", "Xmvvualtokh", "Rbaliuycliu", "Xlhofowltm", "Jqgigdqsvqtw", "Mrvmytfgxyb", "Krwyqoqmeih",
            "Zcucwtks", "Lphobkhmtio", "Gwmtzgbm", "Wryqyvfr", "Koblytwog", "Oqqwmlsoycmg", "Mrplaoqwi"
        ])
        self.assertEqual(length, len(self.taxon), "Filter by all orders.")
        self.taxon.filter_by_order(["Rbaliuycliu", "Xlhofowltm"])
        self.assertEqual(39834 + 4, len(self.taxon), "Incorrect filter.")

    @patch('builtins.__import__', side_effect=import_mock)
    def test_filter_family(self, mock_import):
        self.read_pandas()
        length = len(self.taxon)
        self.assertEqual(length, len(self.taxon), "Original length of dataframe.")
        self.taxon.filter_by_family([
            "Yzpkpquxs", "Aagnljgrt", "Kdahroigttef", "Wbeobsbx", "Uzajhqjglr", "Qmsjwosehsjg", "Jcvvsouvb", "Geitphvl",
            "Szlglzepeoj", "Tuguotkggz", "Csmeldytb", "Yvnmmueju", "Wykuuqdtfh", "Dxojfzqb", "Nwzoybkyquv",
            "Xvflpivgwtw", "Hbznvnpvdlll", "Nytpapprh", "Cdaukfqnbild", "Dvfwfnrosp", "Vowctemwht", "Yhfsclavzwz",
            "Dcohvbddjk", "Rtuitsplmqrw", "Kxwbhmfxwmah", "Abaxaenj", "Cscukzira", "Dxaougdpy", "Kvtgqwfmy",
            "Hoskerxymj", "Bmtbnvhlsn", "Jaksslqkhna", "Gvwkrzuzcsm", "Qejrpechiokg", "Ldbjloyw", "Acxcalcjb",
            "Exudmuohg", "Jisevptt", "Dmlkwsbankpr", "Ebxkmrllsmhb", "Fhhnxbgsj", "Blvgjqyf", "Ofrmsymysbj",
            "Iegtwpmzsn", "Uxlbjpgytrr", "Yppvqkvpyg", "Ufeoonfwvbwj", "Wxnibhsng", "Gzhzmloknffb", "Yqzeqqieric",
            "Bwysxkndbdo", "Aujqbritahm", "Nnjnmwvudn", "Pnpuqusvjmoh", "Qubplgsfgbf", "Aatwuzftuuh", "Mieccdvk",
            "Dakljsno", "Jqvrvbesgygm", "Iuzgldlop", "Xiuuksnsmk", "Cvbllcbq", "Hfwzlwzsmin", "Rqbbjpso", "Ipfvwairk",
            "Nbtktczp", "Jodyeaftwmzy", "Xuejrpuqj", "Xbezapgssp", "Iqrmrnbiovqz", "Wkttduowlm", "Xnxaxfml",
            "Bzivoaubjpig"
        ])
        self.assertEqual(length, len(self.taxon), "Filter by all families.")
        self.taxon.filter_by_family(["Dxaougdpy", "Kvtgqwfmy"])
        self.assertEqual(9969 + 5, len(self.taxon), "Incorrect filter.")

    @patch('builtins.__import__', side_effect=import_mock)
    def test_filter_genus(self, mock_import):
        self.read_pandas()
        length = len(self.taxon)
        self.assertEqual(length, len(self.taxon), "Original length of dataframe.")
        self.taxon.filter_by_genus(["Hbqrtfopjuh", "Vbpjkmfqd"])
        self.assertEqual(624 + 11, len(self.taxon), "Incorrect filter.")

    @patch('builtins.__import__', side_effect=import_mock)
    def test_filter_species(self, mock_import):
        self.read_pandas()
        length = len(self.taxon)
        self.assertEqual(length, len(self.taxon), "Original length of dataframe.")
        species_synonym = "Hqmjhacvb (Azvoxtzhwueu) gupfjmnf"
        variety = "Rtfkiaicpdng obzninluz var. cebwyzcqoy Glhnskwn"
        cultivar = "Rtfkiaicpdng (Abifwvxqn) gurqtwpof f. prczacpvtdtu 'xzhgezqpaorp'"
        self.taxon.filter_by_species([species_synonym, variety, cultivar])
        self.assertEqual(24, len(self.taxon), "Filter by species.")

    @patch('builtins.__import__', side_effect=import_mock)
    def test_filter_species_fuzzy_threshold(self, mock_import):
        self.read_pandas()
        length = len(self.taxon)
        self.assertEqual(length, len(self.taxon), "Original length of dataframe.")
        species_synonym = "Hqmjhacvb (Azvoxtzhwueu) gupfjmnf"
        variety = "Rtfkiaicpdng obzninluz var. cebcqoy Glhnskwn"
        cultivar = "Rtfkiaicpdng (Abifwvxqn) gurqtwpof 'xzhgezqpaorp'"
        self.taxon.filter_by_species([species_synonym, variety, cultivar], fuzzy_threshold=85)
        self.assertEqual(24, len(self.taxon), "Filter by species with enough threshold.")
        self.taxon.filter_by_species([species_synonym, variety, cultivar], fuzzy_threshold=90)
        self.assertGreater(24, len(self.taxon), "Filter by species with not enough threshold.")

    @patch('builtins.__import__', side_effect=import_mock)
    def test_filter_phylum_exception(self, mock_import):
        super().__test_filter_phylum_exception__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_none(self, mock_import):
        super().__test_none__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_merge(self, mock_import):
        mock_import.side_effect = import_mock
        super().__test_merge__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_no_pandas(self, mock_import):
        self.read_pandas()
        with self.assertRaisesRegex(ImportError, "Install pandas to use this feature"):
            var = self.taxon.pandas

    @patch('builtins.__import__', side_effect=import_mock)
    def test_set_core_field(self, mock_import):
        self.__test_set_core_field__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_set_primary_key(self, mock_import):
        self.__test_set_primary_key__()

    @patch('builtins.__import__', side_effect=import_mock)
    def test_sql_table(self, mock_import):
        self.__test_sql_table__()

    @patch('builtins.__import__', side_effect=import_mock_fuzzy)
    def test_fuzzy_exception(self, mock_import):
        self.__test_fuzzy_exception__()


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from typing import Dict, List

import lxml.etree as et

from eml.types import _NoTagObject, EMLTextType


class ProcedureStep(_NoTagObject):
    """
    Represent a description of a specific step of the method.

    Parameters
    ----------
    description : EMLTextType
        Description of the methods employed in collecting or generating a data set or other resource or in quality control and assurance.
    citation : EMLCitation, optional
        Literature citation relating to the methods used.
    protocol : EMLProtocol, optional
        Protocol description relating to the methods used.
    instrumentation : List[str], optional
        Instruments used for measurement and recording data.
    software : List[EMLSoftware], optional
        Software used in the processing of data.
    sub_step : List[ProcedureStep], optional
        This fields allows the nesting of additional method steps within this step.
    """
    def __init__(
            self, description: EMLTextType,
            citation: EMLCitation = None,
            protocol: EMLProtocol = None,
            instrumentation: List[str] = None,
            software: List[EMLSoftware] = None,
            sub_step: List[ProcedureStep] = None
    ) -> None:
        super().__init__()
        self.__description__ = description
        if citation is not None and protocol is not None:
            raise TypeError("ProcedureStep() allow only one of citation (`eml.resource.EMLCitation`) "
                            "or a protocol (`eml.resource.EMLProtocol`).")
        self.__citation__ = citation
        self.__protocol__ = protocol
        self.__instrumentation__ = list()
        if instrumentation is not None:
            self.__instrumentation__.extend(instrumentation)
        self.__software__ = list()
        if software is not None:
            self.__software__.extend(software)
        self.__sub_step__ = list()
        if sub_step is not None:
            self.__sub_step__.extend(sub_step)
        return

    @property
    def description(self) -> EMLTextType:
        """EMLTextType: Description of the methods employed in collecting or generating a data set or other resource or in quality control and assurance."""
        return self.__description__

    @property
    def citation(self) -> EMLCitation:
        """EMLCitation: Literature citation relating to the methods used."""
        return self.__citation__

    @property
    def protocol(self) -> EMLProtocol:
        """EMLProtocol: Protocol description relating to the methods used."""
        return self.__protocol__

    @property
    def instrumentation(self) -> List[str]:
        """List[str]: Instruments used for measurement and recording data."""
        return self.__instrumentation__

    @property
    def software(self) -> List[EMLSoftware]:
        """List[EMLSoftware]: Software used in the processing of data."""
        return self.__software__

    @property
    def sub_step(self) -> List[ProcedureStep]:
        """List[ProcedureStep]: List of sub step, useful for hierarchical method descriptions."""
        return self.__sub_step__

    @classmethod
    def parse(cls, element: et.Element, nmap: Dict) -> ProcedureStep | None:
        """
        Parses a lxml element into a Procedure Step object.

        Parameters
        ----------
        element : lxml.tree.Element
            Object in the Element format to be parsed into a Python object.
        nmap : Dict
            Dictionary of namespace.

        Returns
        -------
        ProcedureStep
            Object parsed.
        """
        if element is None:
            return None
        citation_elem = element.find("citation", nmap)
        if citation_elem is not None:
            from eml.resources import EMLCitation
            citation = EMLCitation.parse(citation_elem, nmap)
        else:
            citation = None
        protocol_elem = element.find("protocol", nmap)
        if protocol_elem is not None:
            from eml.resources import EMLProtocol
            protocol = EMLProtocol.parse(protocol_elem, nmap)
        else:
            protocol = None
        instrumentation = list()
        for instrument_elem in element.findall("instrumentation", nmap):
            instrumentation.append(instrument_elem.text)
        software = list()
        software_list = element.findall("software", nmap)
        if len(software_list) > 0:
            from eml.resources import EMLSoftware
            for software_elem in software_list:
                software.append(EMLSoftware.parse(software_elem, nmap))
        sub_step = list()
        for sub_step_elem in element.findall("subStep", nmap):
            sub_step.append(ProcedureStep.parse(sub_step_elem, nmap))
        return ProcedureStep(
            description=EMLTextType.parse(element.find("description", nmap), nmap),
            citation=citation,
            protocol=protocol,
            instrumentation=instrumentation,
            software=software,
            sub_step=sub_step
        )

    def to_element(self) -> et.Element:
        """
        Generates a lxml.tree.Element from the Procedure Step object.

        Returns
        -------
        lxml.tree.Element
            Methods instance in the Element format.
        """
        element = super().to_element()
        self.description.set_tag("description")
        element.append(self.description.to_element())
        if self.citation is not None:
            element.append(self.citation.to_element())
        if self.protocol is not None:
            element.append(self.protocol.to_element())
        for instrument in self.instrumentation:
            instrument_elem = self.object_to_element("instrumentation")
            instrument_elem.text = instrument
            element.append(instrument_elem)
        for software in self.software:
            element.append(software.to_element())
        for sub_step in self.sub_step:
            sub_step.set_tag("subStep")
            element.append(sub_step.to_element())
        return element



from typing import List, Tuple

from dwca.terms import Field


class OrganismID(Field):
    """
    An identifier for the Organism instance (as opposed to a particular digital record of the Organism).

    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/organismID"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class OrganismName(Field):
    """
    A textual name or label assigned to an Organism instance.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/organismName"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class OrganismScope(Field):
    """
    A description of the kind of Organism instance.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    vocabulary: str, optional
        An URI for a vocabulary that the source values for this Field are based on.
    """
    URI = "http://rs.tdwg.org/dwc/terms/organismScope"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, vocabulary: str = None) -> None:
        super().__init__(index, default, vocabulary)
        return


class AssociatedOrganisms(Field):
    """
    A list of identifiers of other Organisms and the associations of this Organism to each of them.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/associatedOrganisms"
    TYPE = List[Tuple[str, str]]

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class PreviousIdentifications(Field):
    """
    A list of previous assignments of names to the Organism.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/previousIdentifications"
    TYPE = List[str]

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return


class OrganismRemarks(Field):
    """
    Comments or notes about the Organism instance.
    
    Parameters
    ----------
    index : int | str
        Specifies the position of the column in the row.
    default: TYPE, optional
        Specifies a value to use if one is not supplied.
    """
    URI = "http://rs.tdwg.org/dwc/terms/organismRemarks"
    TYPE = str

    def __init__(self, index: int | str, default: TYPE = None, **kwargs) -> None:
        super().__init__(index, default, None)
        return



import os.path
import unittest
from typing import Tuple, Dict

from lxml import etree as et

from dwca.classes import Taxon
from dwca.terms import DWCLanguage, TaxonID
from test_dwca.test_dwca_no_pandas import orig_import
from test_xml.test_xml import TestXML
from xml_common.utils import Language

PATH = os.path.abspath(os.path.dirname(__file__))


class TestTaxonCommon(TestXML):
    @staticmethod
    def read_xml(path: str) -> Tuple[Dict, str, str]:
        with open(path, "r", encoding="utf-8") as file:
            content = file.read()
        base_file = et.fromstring(content)
        nmap = base_file.nsmap
        xml_file = base_file.find("core", namespaces=base_file.nsmap)
        text = et.tostring(xml_file, pretty_print=True).decode("utf-8")
        return nmap, xml_file, text

    def setUp(self) -> None:
        self.nmap, self.taxon_xml, self.text = self.read_xml(os.path.join(PATH, os.pardir, "example_data", "meta.xml"))
        self.taxon = None
        return

    def read_pandas(self, lazy: bool = False):
        self.taxon = Taxon.from_string(self.text)
        with open(os.path.join(PATH, os.pardir, "example_data", "taxon.txt"), "rb") as file:
            if lazy:
                self.taxon.read_file("", source_file=file, lazy=lazy)
            else:
                self.taxon.read_file(file.read().decode(encoding="utf-8"))
        return

    def __test_add_field__(self):
        taxon = Taxon.from_string(self.text)
        taxon.add_field(DWCLanguage(index=47, default=Language.ENG))
        self.assertEqual(48, len(taxon.__fields__), "Fields not set")
        self.text = self.text.replace(
            """<field index="46" default="EXAMPLE" term="http://rs.tdwg.org/dwc/terms/institutionCode"/>""",
            """<field index="46" default="EXAMPLE" term="http://rs.tdwg.org/dwc/terms/institutionCode"/>
    <field index="47" default="eng" term="http://purl.org/dc/elements/1.1/language"/>""")
        self.assertEqualTree(et.fromstring(self.text), taxon.to_element(), "Error on element conversion with new field")

    def __test_add_field_incorrect_index__(self):
        taxon = Taxon.from_string(self.text)
        with self.assertWarnsRegex(UserWarning, "index"):
            taxon.add_field(DWCLanguage(index=3, default=Language.ENG))
        self.assertEqual(48, len(taxon.__fields__), "Fields not set")
        self.text = self.text.replace(
            """<field index="46" default="EXAMPLE" term="http://rs.tdwg.org/dwc/terms/institutionCode"/>""",
            """<field index="46" default="EXAMPLE" term="http://rs.tdwg.org/dwc/terms/institutionCode"/>
    <field index="47" default="eng" term="http://purl.org/dc/elements/1.1/language"/>""")
        self.assertEqualTree(et.fromstring(self.text), taxon.to_element(), "Error on element conversion with new field")

    def __test_parse__(self):
        taxon = Taxon.from_string(self.text)
        self.assertEqual("taxon.txt", taxon.filename, "Files parse incorrectly")
        self.assertEqual(0, taxon.id, "Files parse incorrectly")
        self.assertEqual("UTF-8", taxon.__encoding__, "Encoding parse incorrectly")
        self.assertEqual("\n", taxon.__lines_end__, "Lines end parse incorrectly")
        self.assertEqual("\t", taxon.__fields_end__, "Field end parse incorrectly")
        self.assertEqual("", taxon.__fields_enclosed__, "Field enclosed parse incorrectly")
        self.assertEqual(1, taxon.__ignore_header_lines__, "Ignore header parse incorrectly")
        self.assertEqual(Taxon.URI, taxon.uri, "Row type parse incorrectly")
        self.assertEqual(47, len(taxon.__fields__), "Fields parse incorrectly")
        self.assertEqualTree(self.taxon_xml, taxon.to_element(), "Error on element conversion")

    def __test_filter_kingdom_exception__(self):
        taxon = Taxon(0, "file.txt", [])
        self.assertRaisesRegex(
            AssertionError,
            "Kingdom must be in fields",
            taxon.filter_by_kingdom,
            []
        )

    def __test_get_parents__(self):
        self.read_pandas()
        parents_set = self.taxon.get_parents(["urn:lsid:example.org:taxname:360"])
        self.assertEqual(11, len(parents_set), "Incorrect number of parents.")
        parents_set = self.taxon.get_parents(["urn:lsid:example.org:taxname:366"])
        self.assertEqual(14, len(parents_set), "Incorrect number of parents.")
        parents_set = self.taxon.get_parents(["urn:lsid:example.org:taxname:360", "urn:lsid:example.org:taxname:366"])
        self.assertEqual(14, len(parents_set), "Incorrect number of parents.")

    def __test_get_incorrect_parent__(self):
        self.read_pandas()
        taxa_id = "urn:lsid:example.org:taxname:300000"
        with self.assertWarnsRegex(RuntimeWarning, taxa_id):
            missing = self.taxon.get_parents([taxa_id])
            self.assertEqual(0, len(missing), "Parent found of invalid taxon id")

    def __test_get_synonyms__(self):
        self.read_pandas()
        first_synonyms = self.taxon.all_synonyms(["urn:lsid:example.org:taxname:10000"])
        self.assertEqual(2, len(first_synonyms), "Incorrect number of synonyms.")
        second_synonyms = self.taxon.all_synonyms(["urn:lsid:example.org:taxname:10001"])
        self.assertEqual(2, len(second_synonyms), "Incorrect number of synonyms.")
        self.assertCountEqual(
            first_synonyms, second_synonyms, "Synonyms do not match"
        )
        two_synonyms = self.taxon.all_synonyms(["urn:lsid:example.org:taxname:10000", "urn:lsid:example.org:taxname:100005"])
        self.assertEqual(3, len(two_synonyms), "Incorrect number of synonym with a taxa with no synonymous.")
        self.assertTrue("urn:lsid:example.org:taxname:100005" in two_synonyms, "Incorrect same synonym")

    def __test_get_synonyms_names__(self):
        self.read_pandas()
        first_synonyms = self.taxon.all_synonyms(["urn:lsid:example.org:taxname:10000"], get_names=True)
        self.assertEqual(2, len(first_synonyms), "Incorrect number of synonyms.")
        second_synonyms = self.taxon.all_synonyms(["urn:lsid:example.org:taxname:10001"], get_names=True)
        self.assertEqual(2, len(second_synonyms), "Incorrect number of synonyms.")
        self.assertCountEqual(
            first_synonyms, second_synonyms, "Synonyms do not match"
        )
        two_synonyms = self.taxon.all_synonyms([
            "urn:lsid:example.org:taxname:10000", "urn:lsid:example.org:taxname:100005"
        ], get_names=True)
        self.assertEqual(3, len(two_synonyms), "Incorrect number of synonym with a taxa with no synonymous.")
        self.assertTrue(
            "Wxjncmbdqbd (Jdhbtbju) oojmaovly var. ushibyxcfc Whhngbhiq" in two_synonyms,
            "Incorrect same synonym"
        )

    def __test_get_incorrect_synonyms__(self):
        self.read_pandas()
        taxa_id = "urn:lsid:example.org:taxname:300000"
        with self.assertWarnsRegex(RuntimeWarning, "urn:lsid:example.org:taxname:300000"):
            synonyms = self.taxon.all_synonyms([taxa_id])
            self.assertEqual(0, len(synonyms), "Found something of false taxon id.")

    def __test_filter_phylum_exception__(self):
        taxon = Taxon(0, "file.txt", [])
        self.assertRaisesRegex(
            AssertionError,
            "Phylum must be in fields",
            taxon.filter_by_phylum,
            []
        )

    def __test_fuzzy_exception__(self):
        self.read_pandas()
        self.assertRaisesRegex(
            ImportError,
            "Install rapidfuzz to use this feature.",
            self.taxon.filter_by_species,
            [],
            fuzzy_threshold=90
        )

    def __test_none__(self):
        self.assertIsNone(Taxon.parse(None, {}), "Object parsed from nothing")

    def __test_merge__(self):
        _, xml_file_1, text_1 = self.read_xml(os.path.join(PATH, os.pardir, "example_data", "dataset1_meta.xml"))
        _, xml_file_2, text_2 = self.read_xml(os.path.join(PATH, os.pardir, "example_data", "dataset2_meta.xml"))
        taxon1 = Taxon.from_string(text_1)
        with open(os.path.join(PATH, os.pardir, "example_data", "taxon1.txt"), "r", encoding="utf-8") as taxa_file:
            taxon1.read_file(taxa_file.read())
        taxon2 = Taxon.from_string(text_2)
        with open(os.path.join(PATH, os.pardir, "example_data", "taxon2.txt"), "r", encoding="utf-8") as taxa_file:
            taxon2.read_file(taxa_file.read())
        merged_taxon = Taxon.merge(taxon1, taxon2)
        self.assertEqual(len(taxon1) + len(taxon2), len(merged_taxon), "Incorrect merged of taxa files.")
        self.assertCountEqual(
            merged_taxon.fields,
            taxon1.fields,
            "Incorrect number of fields"
        )
        self.assertCountEqual(
            merged_taxon.fields,
            taxon2.fields,
            "Incorrect number of fields"
        )

    def __test_set_core_field__(self):
        self.read_pandas()
        self.assertRaisesRegex(
            AttributeError,
            "Core",
            self.taxon.set_core_field,
            TaxonID(0)
        )

    def __test_set_primary_key__(self):
        self.read_pandas()
        self.assertRaisesRegex(
            AttributeError,
            "Core",
            self.taxon.set_primary_key,
            "Taxon"
        )

    def __test_sql_table__(self):
        self.read_pandas()
        self.assertEqual(self.normalize_sql("""
        CREATE TABLE "Taxon" (
            "taxonID" VARCHAR,
            "scientificNameID" VARCHAR,
            "acceptedNameUsageID" VARCHAR,
            "parentNameUsageID" VARCHAR,
            "namePublishedInID" VARCHAR,
            "originalNameUsageID" VARCHAR,
            "nameAccordingToID" VARCHAR,
            "taxonConceptID" VARCHAR,
            "scientificName" VARCHAR,
            "acceptedNameUsage" VARCHAR,
            "parentNameUsage" VARCHAR,
            "namePublishedIn" VARCHAR,
            "originalNameUsage" VARCHAR,
            "nameAccordingTo" VARCHAR,
            "namePublishedInYear" INTEGER,
            "higherClassification" VARCHAR,
            "kingdom" VARCHAR,
            "phylum" VARCHAR,
            "class" VARCHAR,
            "order" VARCHAR,
            "superfamily" VARCHAR,
            "family" VARCHAR,
            "subfamily" VARCHAR,
            "tribe" VARCHAR,
            "subtribe" VARCHAR,
            "genus" VARCHAR,
            "genericName" VARCHAR,
            "subgenus" VARCHAR,
            "infragenericEpithet" VARCHAR,
            "specificEpithet" VARCHAR,
            "infraspecificEpithet" VARCHAR,
            "cultivarEpithet" VARCHAR,
            "taxonRank" VARCHAR,
            "verbatimTaxonRank" VARCHAR,
            "scientificNameAuthorship" VARCHAR,
            "vernacularName" VARCHAR,
            "nomenclaturalCode" VARCHAR,
            "taxonomicStatus" VARCHAR,
            "nomenclaturalStatus" VARCHAR,
            "taxonRemarks" VARCHAR,
            "modified" VARCHAR,
            "bibliographicCitation" VARCHAR,
            "references" VARCHAR,
            "license" VARCHAR,
            "rightsHolder" VARCHAR,
            "datasetName" VARCHAR,
            "institutionCode" VARCHAR,
            PRIMARY KEY ("taxonID")
        );
        """), self.normalize_sql(self.taxon.sql_table), "Wrong SQL table.")


if __name__ == '__main__':
    unittest.main()



import unittest
import datetime as dt
from typing import Union

from xml_common.utils import type_to_sql, format_to_sql


class TestTypeSQL(unittest.TestCase):
    def test_int(self):
        self.assertEqual("INTEGER", type_to_sql(int), "Error converting int.")
        self.assertEqual(5, format_to_sql(5, int), "Error formatting to int.")
        self.assertEqual(5, format_to_sql("5", int), "Error formatting to int, from string.")
        self.assertIsNone(format_to_sql(None, int), "Error formatting to int, from None.")

    def test_float(self):
        self.assertEqual("REAL", type_to_sql(float), "Error converting float.")
        self.assertEqual(0.5, format_to_sql(0.5, float), "Error formatting to float.")
        self.assertEqual(1, format_to_sql(1, float), "Error formatting to float, from string.")
        self.assertIsNone(format_to_sql(None, float), "Error formatting to float, from None.")

    def test_str(self):
        self.assertEqual("VARCHAR", type_to_sql(str), "Error converting str.")
        self.assertEqual("1", format_to_sql(1, str), "Error formatting to string.")
        self.assertEqual("Hello", format_to_sql("Hello", str), "Error formatting to string.")

    def test_bool(self):
        self.assertEqual("BOOLEAN", type_to_sql(bool), "Error converting bool.")
        self.assertTrue(format_to_sql(True, bool), "Error formatting to bool.")
        self.assertFalse(format_to_sql(False, bool), "Error formatting to bool.")

    def test_datetime(self):
        self.assertEqual("DATETIME", type_to_sql(dt.datetime), "Error converting datatime.")
        a_date = dt.datetime(1997, 10, 3)
        self.assertEqual(a_date, format_to_sql(a_date, dt.datetime), "Error formatting to datetime.")
        self.assertEqual(a_date, format_to_sql("1997-10-03", dt.datetime), "Error formatting to datetime, from string.")
        with self.assertRaisesRegex(ValueError, "is not supported to datetime"):
            format_to_sql(1997, dt.datetime)

    def test_bytes(self):
        self.assertEqual("BLOB", type_to_sql(bytes), "Error converting byte.")
        self.assertEqual(b"Hello", format_to_sql(b"Hello", bytes), "Error formatting to bytes.")
        self.assertEqual(b"Hello", format_to_sql("Hello", bytes), "Error formatting to bytes, from string.")
        self.assertEqual(bytes(20), format_to_sql(20, bytes), "Error formatting to bytes, from int.")

    def test_object(self):
        self.assertEqual("VARCHAR", type_to_sql(Union[str, int]), "Error converting union.")
        class A:
            pass
        a = A()
        self.assertEqual("VARCHAR", type_to_sql(A), "Error converting object.")
        self.assertEqual(str(a), format_to_sql(a, A), "Error formatting to object.")


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

from typing import Dict
from warnings import warn

from lxml import etree as et

from eml.resources.coverage import GeographicCoverage, TemporalCoverage, TaxonomicCoverage
from eml.types import EMLObject, Scope


class EMLCoverage(EMLObject):
    """
    Extent of the coverage of the resource.

    Parameters
    ----------
    geographic: GeographicCoverage
        Geographic coverage information.
    temporal: TemporalCoverage
        Temporal coverage information.
    taxonomic: TaxonomicCoverage
        Taxonomic coverage information.
    _id : str, optional
        Unique identifier within the scope.
    scope : Scope, default DOCUMENT
        The scope of the identifier.
    system : str, optional
        The data management system within which an identifier is in scope and therefore unique.
    referencing : bool, optional, default=False
        Whether the resource is referencing another or is being defined.
    references_system : str, optional
        System attribute of reference.
    """
    PRINCIPAL_TAG = "coverage"

    def __init__(
            self,
            geographic: GeographicCoverage = None,
            temporal: TemporalCoverage = None,
            taxonomic: TaxonomicCoverage = None,
            _id: str = None,
            scope: Scope = Scope.DOCUMENT,
            system: str = None,
            referencing: bool = False,
            references_system: str = None
    ) -> None:
        super().__init__(_id, scope, system, referencing, references_system)
        if self.referencing:
            return
        if (
            geographic is None and
            temporal is None and
            taxonomic is None
        ):
            raise TypeError("EMLCoverage() required at least one of geographic, temporal and taxonomic coverages")
        self.__geo__ = geographic
        self.__time__ = temporal
        self.__taxa__ = taxonomic
        return

    @property
    def geographic(self) -> GeographicCoverage:
        """GeographicCoverage: Geographic coverage information."""
        return self.__geo__

    @property
    def temporal(self) -> TemporalCoverage:
        """TemporalCoverage: Temporal coverage information."""
        return self.__time__

    @property
    def taxonomic(self) -> TaxonomicCoverage:
        """TaxonomicCoverage: Taxonomic coverage information."""
        return self.__taxa__

    @classmethod
    def get_referrer(cls, element: et.Element, nmap: Dict) -> EMLCoverage:
        """
        Generate an EML Coverage referencing another EML Coverage.

        Parameters
        ----------
        element : lxml.etree.Element
            XML element to parse with references object.
        nmap : Dict
            Namespace.

        Returns
        -------
        EMLCoverage
            Object parsed that reference another.
        """
        references = element.find("references", nmap)
        return EMLCoverage(
            _id=references.text if references.text is not None else "",
            scope=cls.get_scope(element),
            system=element.get("system", None),
            referencing=True,
            references_system=references.get("system", None),
        )

    @classmethod
    def get_no_referrer(cls, element: et.Element, nmap: Dict) -> EMLCoverage:
        """
        Generate an EML Coverage that do not reference another.

        Parameters
        ----------
        element : lxml.etree.Element
            XML element to parse.
        nmap : Dict
            Namespace.

        Returns
        -------
        EMLCoverage
            Object parsed.
        """
        try:
            geographic = GeographicCoverage.parse(element.find("geographicCoverage", nmap), nmap)
        except ValueError as e:
            warn(str(e), category=SyntaxWarning)
            geographic = None
        try:
            temporal = TemporalCoverage.parse(element.find("temporalCoverage", nmap), nmap)
        except ValueError as e:
            warn(str(e), category=SyntaxWarning)
            temporal = None
        try:
            taxonomic = TaxonomicCoverage.parse(element.find("taxonomicCoverage", nmap), nmap)
        except ValueError as e:
            warn(str(e), category=SyntaxWarning)
            taxonomic = None
        return EMLCoverage(
            geographic=geographic,
            temporal=temporal,
            taxonomic=taxonomic,
            _id=element.get("id", None),
            scope=cls.get_scope(element),
            system=element.get("system", None),
            referencing=False,
        )

    def to_element(self) -> et.Element:
        """
        Generate an XML element instance.

        Returns
        -------
        lxml.etree.Element
            XML element instance.
        """
        element = super().to_element()
        element = self._to_element_(element)
        if not self.referencing:
            element.append(self.geographic.to_element()) if self.geographic is not None else None
            element.append(self.temporal.to_element()) if self.temporal is not None else None
            element.append(self.taxonomic.to_element()) if self.taxonomic is not None else None
        return element



import unittest
from unittest.mock import patch

import tqdm

from xml_common.utils import OptionalTqdm, iterate_with_bar
from xml_common.utils.iteratate import is_notebook, get_optional_iterator

orig_import = __import__

def import_mock(name, globals=None, locals=None, fromlist=(), level=0):
    if name == 'IPython':
        raise ImportError(f"No module named '{name}'")
    return orig_import(name, globals, locals, fromlist, level)


def import_mock_tqdm(name, globals=None, locals=None, fromlist=(), level=0):
    if name == 'tqdm':
        raise ImportError(f"No module named '{name}'")
    return orig_import(name, globals, locals, fromlist, level)


class TestIterator(unittest.TestCase):
    @patch("IPython.get_ipython", return_value=True)
    def test_is_notebook(self, mock_get_ipython):
        self.assertTrue(is_notebook())

    def test_is_terminal(self):
        self.assertFalse(is_notebook())

    @patch('builtins.__import__', side_effect=import_mock)
    def test_is_terminal_no_module(self, mock_import):
        self.assertFalse(is_notebook())

    @patch('builtins.__import__', side_effect=import_mock_tqdm)
    def test_optional_iterator(self, mock_import):
        a = [1, 2, 6, 5]
        tqdm_optional = get_optional_iterator(iterator=a)
        self.assertEqual(a, tqdm_optional, "Incorrect iterator")
        new_tqdm = get_optional_iterator()
        self.assertIsNone(new_tqdm, "New tqdm is not None")
        iterate_bar = iterate_with_bar(a)
        self.assertEqual(a, tqdm_optional, "Incorrect bar iterator")

    def test_optional_tqdm_iterator(self):
        a = [1, 2, 6, 5]
        tqdm_optional = get_optional_iterator(iterator=a)
        self.assertTrue(isinstance(tqdm_optional, tqdm.tqdm), "Incorrect iterator")
        new_tqdm = get_optional_iterator()
        self.assertTrue(isinstance(new_tqdm, tqdm.tqdm), "Incorrect new tqdm")
        iterate_bar = iterate_with_bar(a)
        self.assertTrue(isinstance(iterate_bar, tqdm.tqdm), "Incorrect bar iterator")

    @patch("IPython.get_ipython", return_value=True)
    def test_optional_tqdm_ipython(self, mock_get_ipython):
        a = [1, 2, 6, 5]
        # Strange behaviour because of this is not a notebook
        tqdm_optional = get_optional_iterator(iterator=a)
        self.assertEqual(a, tqdm_optional, "Incorrect iterator")
        new_tqdm = get_optional_iterator()
        self.assertIsNone(new_tqdm, "New tqdm is not None")
        tqdm_optional = iterate_with_bar(a)
        self.assertEqual(a, tqdm_optional, "Incorrect bar iterator")

    def test_optional_tqdm(self):
        a = OptionalTqdm()
        self.assertTrue(isinstance(a.__tqdm__, tqdm.tqdm), "Incorrect tqdm")
        a.set_descriptor(desc="A descriptor")
        self.assertEqual("A descriptor: ", a.__tqdm__.desc, "Descriptor not set")
        self.assertEqual(0, a.__tqdm__.n, "Wrong initial")
        a.update()
        self.assertEqual(1, a.__tqdm__.n, "Not updated")
        a.reset()
        self.assertEqual(0, a.__tqdm__.n, "Not reset")
        self.assertIsNone(a.__tqdm__.postfix, "Postfix initialized")
        a.set_postfix()
        self.assertIsNotNone(a.__tqdm__.postfix, "Postfix not set")
        self.assertEqual("", a.__tqdm__.postfix, "Postfix not set")
        self.assertFalse(a.__tqdm__.disable, "Disable")
        a.close()
        self.assertTrue(a.__tqdm__.disable, "Not disable")

    @patch("builtins.__import__", side_effect=import_mock_tqdm)
    def test_optional_tqdm_none(self, mock_import):
        a = OptionalTqdm()
        self.assertIsNone(a.__tqdm__, "Incorrect tqdm (not none)")
        # Assert not raise any error
        a.set_descriptor(desc="A descriptor")
        a.update()
        a.reset()
        a.set_postfix()
        a.close()
        a.set_postfix()


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

from typing import Dict

from lxml import etree as et

from eml.resources import Resource


class EMLCitation(Resource):
    """
    EML Citation Resource.

    Other Parameters
    ----------------
    **kwargs : :class:`eml.resources.resource.Resource` parameters.
        The parameters of every type of Resource.
    """
    PRINCIPAL_TAG = "citation"

    def __init__(
            self, **kwargs
    ) -> None:
        super().__init__(**kwargs)
        return

    # TODO: Implement citation attributes (https://eml.ecoinformatics.org/schema/eml_xsd#eml_citation)

    @classmethod
    def get_referrer(cls, element: et.Element, nmap: Dict) -> EMLCitation:
        """
        Generate an EML Citation referencing another EML Citation.

        Parameters
        ----------
        element : lxml.etree.Element
            XML element to parse with references another citation.
        nmap : Dict
            Namespace.

        Returns
        -------
        EMLCitation
            Object parsed that reference another citation.
        """
        references = element.find("references", nmap)
        return EMLCitation(
            _id=references.text,
            scope=cls.get_scope(element),
            system=element.get("system", None),
            referencing=True,
            references_system=references.get("system", None)
        )

    @classmethod
    def get_no_referrer(cls, element: et.Element, nmap: Dict) -> EMLCitation:
        """
        Generate an EML citation that do not reference another citation.

        Parameters
        ----------
        element : lxml.etree.Element
            XML element to parse.
        nmap : Dict
            Namespace.

        Returns
        -------
        EMLCitation
            Object parsed.
        """
        pass

    def to_element(self) -> et.Element:
        """
        Generate an XML `Element` from this object

        Returns
        -------
        lxml.etree.Element
            XML `Element` from this object
        """
        cite_elem = super().to_element()
        cite_elem = self._to_element_(cite_elem)
        return cite_elem



import unittest

from lxml import etree as et

from eml.resources.distribution import EMLInline
from test_xml.test_xml import TestXML


class TestInline(TestXML):
    def test_parse(self):
        text_xml = """
<inline></inline>
        """
        inline = EMLInline.from_string(text_xml)
        self.assertEqualTree(et.fromstring(text_xml), inline.to_element(), "Error on to element")

    def test_parse_none(self):
        self.assertIsNone(EMLInline.parse(None, {}), "EML Inline from nowhere")


if __name__ == '__main__':
    unittest.main()
